"""Group of functions for physiological processing."""
from re import findall
import logging

import numpy as np
from itertools import product
from scipy.stats import zscore

from .sigproc import smoothing

__all__ = ('find_non_eeg', 'rereferencing', 'bipolarization', 'commonaverage',
           'tal2mni', 'mni2tal', 'generate_eeg')

logger = logging.getLogger('visbrain')


def find_non_eeg(channels, pattern=['eog', 'emg', 'ecg', 'abd']):
    """Find non-EEG channels.

    Parameters
    ----------
    channels : list
        List of channel names.
    pattern : list | ['eog', 'emg', 'ecg', 'abd']
        List of patterns for non-EEG channels.

    Returns
    -------
    iseeg : array_like
        NumPy vector of boolean values.
    """
    # Set channels in lower case :
    channels = np.char.lower(np.asarray(channels))
    # Pre-allocation :
    iseeg = np.zeros((len(channels),), dtype=bool)
    # Search patterns :
    for k in pattern:
        iseeg += np.invert(np.char.find(channels, k).astype(bool))
    return iseeg


###############################################################################
###############################################################################
#                               RE-REFERENCING
###############################################################################
###############################################################################

def rereferencing(data, chans, reference, to_ignore=None):
    """Re-reference data.

    Parameters
    ----------
    data : array_like
        The array of data of shape (nchan, npts).
    chans : list
        List of channel names of length nchan.
    reference : int
        The index of the channel to consider as a reference.
    to_ignore : list | None
        List of channels to ignore in the re-referencing.

    Returns
    -------
    datar : array_like
        The re-referenced data.
    channelsr : list
        List of re-referenced channel names.
    consider : list
        List of boolean values of channels that have to be considered
        during the ploting processus.
    """
    # Get shapes :
    nchan, npts = data.shape
    # Get data to use as the reference :
    ref = data[[reference], :]
    name = chans[reference]
    # Build ignore vector :
    consider = np.ones((nchan,), dtype=bool)
    consider[reference] = False
    # Find if some channels have to be ignored :
    if to_ignore is None:
        sl = slice(nchan)
    elif isinstance(to_ignore, (tuple, list, np.ndarray)):
        to_ignore = np.asarray(to_ignore)
        sl = np.arange(nchan)[~to_ignore]
        consider[to_ignore] = False
    # Re-reference data :
    data[sl, :] -= ref
    # Build channel names :
    chan = [k + '-' + name if consider[num]
            else k for num, k in enumerate(chans)]

    return data, chan, consider


def bipolarization(data, chans, to_ignore=None, sep='.'):
    """Bipolarize data.

    Parameters
    ----------
    data : array_like
        The array of data of shape (nchan, npts).
    chans : list
        List of channel names of length nchan.
    to_ignore : list | None
        List of channels to ignore in the bipolarization.
    sep : string | '.'
        Separator to simplify electrode names by removing undesired name
        after the sep. For example, if channel = ['h1.025', 'h2.578']
        and sep='.', the final name will be 'h2-h1'.

    Returns
    -------
    datar : array_like
        The re-referenced data.
    channelsr : list
        List of re-referenced channel names.
    consider : list
        List of boolean values of channels that have to be considered
        during the ploting processus.
    """
    # Variables :
    nchan, npts = data.shape
    consider = np.ones((nchan,), dtype=bool)

    # Preprocess channel names by separating channel names / number:
    chnames, chnums = [], []
    for num, k in enumerate(chans):
        # Remove spaces and separation :
        chans[num] = k.strip().replace(' ', '').split(sep)[0]
        # Get only the name / number :
        if findall(r'\d+', k):
            number = findall(r'\d+', k)[0]
            chnums.append(number)
            chnames.append(k.split(number)[0])
        else:
            chnums.append('')
            chnames.append(k)

    # Find if some channels have to be ignored :
    if to_ignore is None:
        sl = range(nchan)
    elif isinstance(to_ignore, (tuple, list, np.ndarray)):
        to_ignore = np.asarray(to_ignore)
        sl = np.arange(nchan)[~to_ignore]
        consider[to_ignore] = False

    # Bipolarize :
    for num in reversed(range(nchan)):
        # If there's a number :
        if chnums[num] and (num in sl):
            # Get the name of the channel to find :
            chan_to_find = chnames[num] + str(int(chnums[num]) - 1)
            # Search if exist in channel list :
            if chan_to_find in chans:
                # Get the index :
                ind = chans.index(chan_to_find)
                # Substract to data :
                data[num, :] -= data[ind, :]
                # Update channel name :
                chans[num] = chans[num] + '-' + chan_to_find
            else:
                consider[num] = False
        else:
            consider[num] = False

    return data, chans, consider


def commonaverage(data, chans, to_ignore=None):
    """Re-referencement using common average.

    Parameters
    ----------
    data : array_like
        The array of data of shape (nchan, npts).
    chans : list
        List of channel names of length nchan.
    to_ignore : list | None
        List of channels to ignore in the re-referencing.

    Returns
    -------
    datar : array_like
        The re-referenced data.
    channelsr : list
        List of re-referenced channel names.
    consider : list
        List of boolean values of channels that have to be considered
        during the ploting processus.
    """
    # Variables :
    nchan, npts = data.shape
    consider = np.ones((nchan,), dtype=bool)
    # Find if some channels have to be ignored :
    if to_ignore is not None:
        consider[to_ignore] = False
    # Get the mean across  EEG channels :
    eegmean = data[consider].mean(0, keepdims=True)
    # Remove the mean on EEG channels :
    data[consider, :] -= eegmean
    # Update channel name :
    for k in range(len(chans)):
        chans[k] = chans[k] + '-m' if consider[k] else chans[k]
    return data, chans, consider


###############################################################################
###############################################################################
#                               XYZ CONVERSION
###############################################################################
###############################################################################

def _spm_matrix(p):
    """Matrix transformation.

    Parameters
    ----------
    p : array_like
        Vector of floats for defining each tranformation. p must be a vector of
        length 9.

    Returns
    -------
    Pr : array_like
        The tranformed array.
    """
    q = [0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0]
    p.extend(q[len(p):12])

    # Translation t :
    t = np.array([[1, 0, 0, p[0]],
                  [0, 1, 0, p[1]],
                  [0, 0, 1, p[2]],
                  [0, 0, 0, 1]])
    # Rotation 1 :
    r1 = np.array([[1, 0, 0, 0],
                   [0, np.cos(p[3]), np.sin(p[3]), 0],
                   [0, -np.sin(p[3]), np.cos(p[3]), 0],
                   [0, 0, 0, 1]])
    # Rotation 2 :
    r2 = np.array([[np.cos(p[4]), 0, np.sin(p[4]), 0],
                   [0, 1, 0, 0],
                   [-np.sin([p[4]]), 0, np.cos(p[4]), 0],
                   [0, 0, 0, 1]])
    # Rotation 3 :
    r3 = np.array([[np.cos(p[5]), np.sin(p[5]), 0, 0],
                   [-np.sin(p[5]), np.cos(p[5]), 0, 0],
                   [0, 0, 1, 0],
                   [0, 0, 0, 1]])
    # Translation z :
    z = np.array([[p[6], 0, 0, 0],
                  [0, p[7], 0, 0],
                  [0, 0, p[8], 0],
                  [0, 0, 0, 1]])
    # Translation s :
    s = np.array([[1, p[9], p[10], 0],
                  [0, 1, p[11], 0],
                  [0, 0, 1, 0],
                  [0, 0, 0, 1]])
    return np.linalg.multi_dot([t, r1, r2, r3, z, s])


def tal2mni(xyz):
    """Transform Talairach coordinates into MNI.

    Parameters
    ----------
    xyz : array_like
        Array of Talairach coordinates of shape (n_sources, 3)

    Returns
    -------
    xyz_r : array_like
        Array of MNI coordinates of shape (n_sources, 3)
    """
    # Check xyz to be (n_sources, 3) :
    if (xyz.ndim != 2) or (xyz.shape[1] != 3):
        raise ValueError("The shape of xyz must be (N, 3).")
    n_sources = xyz.shape[0]

    # Transformation matrices, different zooms above/below AC :
    rotn = np.linalg.inv(_spm_matrix([0., 0., 0., .05]))
    upz = np.linalg.inv(_spm_matrix([0., 0., 0., 0., 0., 0., .99, .97, .92]))
    downz = np.linalg.inv(_spm_matrix([0., 0., 0., 0., 0., 0., .99, .97, .84]))

    # Apply rotation and translation :
    xyz = np.dot(rotn, np.c_[xyz, np.ones((n_sources, ))].T)
    tmp = np.array(xyz)[2, :] < 0.
    xyz[:, tmp] = np.dot(downz, xyz[:, tmp])
    xyz[:, ~tmp] = np.dot(upz, xyz[:, ~tmp])
    return np.array(xyz[0:3, :].T)


def mni2tal(xyz):
    """Transform MNI coordinates into Talairach.

    Parameters
    ----------
    xyz : array_like
        Array of MNI coordinates of shape (n_sources, 3)

    Returns
    -------
    xyz_r : array_like
        Array of Talairach coordinates of shape (n_sources, 3)
    """
    # Check xyz to be (n_sources, 3) :
    if (xyz.ndim != 2) or (xyz.shape[1] != 3):
        raise ValueError("The shape of xyz must be (N, 3).")
    n_sources = xyz.shape[0]

    # Transformation matrices, different zooms above/below AC :
    up_t = _spm_matrix([0., 0., 0., .05, 0., 0., .99, .97, .92])
    down_t = _spm_matrix([0., 0., 0., .05, 0., 0., .99, .97, .84])
    xyz = np.c_[xyz, np.ones((n_sources, ))].T

    tmp = np.array(xyz)[2, :] < 0.
    xyz[:, tmp] = np.dot(down_t, xyz[:, tmp])
    xyz[:, ~tmp] = np.dot(up_t, xyz[:, ~tmp])
    return np.array(xyz[0:3, :].T)


def generate_eeg(sf=512., n_pts=1000, n_channels=1, n_trials=1, n_sines=100,
                 f_min=.5, f_max=160., smooth=50, noise=10, random_state=0):
    """Generate random eeg signals.

    Parameters
    ----------
    sf : float | 512.
        The sampling frequency
    n_pts : int | 1000
        The number of time points.
    n_channels : int | 1
        Number of channels
    n_trials : int | 1
        Number of trials
    n_sines : int | 100
        Number of sines composing each epoch.
    f_min : float | .5
        Minimum frequency for sines.
    f_max : float | 160.
        Maximum frequency for sines.
    smooth : float | 50.
        The smoothing factor. Use larger smoothing to reduce high frequencies.
    noise : float | 10.
        Noise level.
    random_state : int | 0
        Fix the random state for the reproducibility.

    Returns
    -------
    data : array_like
        Dataset as a (n_channels, n_trials, n_pts) array.
    time : array_like
        A (n_pts,) vector containing time values.
    """
    _rnd = np.random.RandomState(random_state)
    n_pts += 100  # edge effect compensation
    signal = np.zeros((n_channels, n_trials, n_pts), dtype=float)
    time = np.arange(n_pts).reshape(-1, 1) / sf
    f_sines = np.linspace(f_min, f_max, num=n_sines, endpoint=True)
    phy = _rnd.uniform(0., 2. * np.pi, (n_pts, n_sines))
    sines = np.sin(2. * np.pi * f_sines.reshape(1, -1) * time + phy)
    amp_log = np.logspace(0, 1, n_sines, base=.1)

    for k, i in product(range(n_channels), range(n_trials)):
        amp = amp_log * _rnd.normal(0., 1., n_sines)
        sig = smoothing(np.dot(sines, amp), smooth, 'hanning')
        sig += _rnd.randn(*sig.shape) / (noise * sig.std())
        signal[k, i] = sig
    signal = zscore(signal, -1)
    signal = signal[..., 50:-50]
    time = time[50:-50]
    return np.squeeze(signal), time



"""Enable to annotate a Sleep file."""

from PyQt5 import QtWidgets
import numpy as np


class UiAnnotate(object):
    """Interactions with annotations."""

    def __init__(self):
        """Init."""
        # Add/remove line :
        self._AnnotateAdd.clicked.connect(self._fcn_annotate_add)
        self._AnnotateRm.clicked.connect(self._fcn_annotate_rm)
        self._AnnotateTable.itemSelectionChanged.connect(
            self._fcn_annotate_goto)

    def _fcn_annotate_add(self, _, xlim=None, txt="Enter your annotation"):
        """Add a ligne to the annotation table."""
        # Add a ligne :
        self._AnnotateTable.setRowCount(self._AnnotateTable.rowCount() + 1)
        rw = self._AnnotateTable.rowCount() - 1
        if xlim is None:
            # Get the current window :
            val = self._SlVal.value()
            step = self._SigSlStep.value()
            win = self._SigWin.value()
            xlim = (val * step, val * step + win)
        # Automatically set the window :
        self._AnnotateTable.setItem(rw, 0, QtWidgets.QTableWidgetItem(
            str(xlim[0])))
        self._AnnotateTable.setItem(rw, 1, QtWidgets.QTableWidgetItem(
            str(xlim[1])))
        # Select the text item :
        self._AnnotateTable.setCurrentCell(rw, 2)
        item = QtWidgets.QTableWidgetItem()
        item.setText(txt)
        self._AnnotateTable.setItem(rw, 2, item)
        self._AnnotateTable.editItem(self._AnnotateTable.item(rw, 2))
        # Send marker annotation to the time-axis :
        self._annot_mark = np.append(self._annot_mark, np.array(xlim).mean())
        self._fcn_slider_move()

    def _fcn_annotate_rm(self):
        """Remove a line to the annotation table."""
        row = self._AnnotateTable.currentRow()
        self._AnnotateTable.removeRow(row)
        self._annot_mark = np.delete(self._annot_mark, row, 0)
        self._fcn_slider_move()

    def _fcn_annotate_goto(self):
        """Go to the annotation location."""
        row = self._AnnotateTable.currentRow()
        if row >= 0:
            sta = float(str(self._AnnotateTable.item(row, 0).text()))
            self._SlGoto.setValue(sta)



"""Visual class for 3-D pictures.

Authors: Etienne Combrisson <e.combrisson@gmail.com>

License: BSD (3-clause)
"""
import numpy as np

from vispy import gloo, visuals, scene

from ..utils import cmap_to_glsl, vispy_array

__all__ = ('PicMesh')

VERT_SHADER = """
#version 120

varying vec4 v_color;
varying float v_data;

void main() {
    v_data = $a_data;
    v_color = $color;
    gl_Position = $transform(vec4($a_position, 1.));
}
"""

FRAG_SHADER = """
#version 120

varying vec4 v_color;

void main() {
    gl_FragColor = vec4(v_color.rgb, $u_alpha);
}
"""


class PicVisual(visuals.Visual):
    """Create a VisPy compatible object for multiple small pictures."""

    def __len__(self):
        """Return the number of sources."""
        return self.n

    def __init__(self, data, pos, width=1., height=1., dxyz=(0., 0., 0.,),
                 select=None, alpha=1., **kwargs):
        """Init."""
        # self.unfreeze()
        self.w = width
        self.h = height
        self._pos = pos
        self._dxyz = np.array(dxyz)
        self.camera = []

        visuals.Visual.__init__(self, VERT_SHADER, FRAG_SHADER)

        # Select pictures :
        if isinstance(select, (list, np.ndarray)):
            data = data[select, ...]
            pos = pos[select, ...]

        # Check data and get vertices and faces :
        self._check_data(data, pos)

        # Get vertices and faces :
        a_position = self._data_to_pos(pos)
        faces, grid = self._get_index()

        # Define index and position buffers :
        self._index_buffer = gloo.IndexBuffer(faces)
        self._pos_buffer = gloo.VertexBuffer(a_position)
        self.shared_program.vert['a_position'] = self._pos_buffer

        # Re-order data :
        self._data = data.ravel()[np.argsort(grid.ravel())]
        self._data_buffer = gloo.VertexBuffer(vispy_array(self._data))
        self.shared_program.vert['a_data'] = self._data_buffer
        # Define the color buffer :
        self.shared_program.frag['u_alpha'] = alpha
        self.alpha = alpha
        self.set_data(**kwargs)

        # Define drawing mode :
        self._draw_mode = 'triangles'

    def _prepare_transforms(self, view):
        """Prepare transformation."""
        tr = view.transforms
        transform = tr.get_transform()

        view_vert = view.view_program.vert
        view_vert['transform'] = transform

    def _check_data(self, data, pos):
        """Check data and position.

        Parameters
        ----------
        data : array_like
            Array of data of shape (n_sources, n_rows, n_cols). If data has
            4 dims, the last one is considered as RGBA.

        pos : array_like
            Position of each center of shape (n_centers, 3).
        """
        # Check position :
        if pos.ndim != 2:
            raise ValueError("The pos variable must be (n_sources, 3)")
        self.n = pos.shape[0]
        # Check data :
        if (data.shape[0] != len(self)) or (data.ndim < 3):
            raise ValueError("The data variable must be (n_sources, n_rows, "
                             "ncols)")
        self.nrows = data.shape[1]
        self.ncols = data.shape[2]

    def _data_to_pos(self, pos):
        """Convert pos into a compatible grid of positions.

        Parameters
        ----------
        pos : array_like
            Position of each center of shape (n_centers, 3).

        Returns
        -------
        grid : array_like
            The grid of positions of shape (n_centers * n_rows * n_cols, 3).
        """
        grid = np.zeros((self.n, self.ncols, self.nrows, 3), dtype=np.float32)
        # Build (starting, ending) indices :
        xg, yg = np.mgrid[-self.w / 2:self.w / 2:self.ncols * 1j,
                          -self.h / 2:self.h / 2:self.nrows * 1j]
        xg = np.flipud(xg)
        for k in range(self.n):
            grid[k, :, :, 0] = pos[k, 0] + xg + self._dxyz[0]
            grid[k, :, :, 1] = pos[k, 1] + yg + self._dxyz[1]
            grid[k, :, :, 2] = pos[k, 2] + self._dxyz[2]
        return grid.reshape(self.n * self.nrows * self.ncols, 3)

    def _get_index(self):
        """Build the index of triangles.

        Returns
        -------
        index : array_like
            Array of indices for the triangles of shape
            (n_centers * 2 * (n_rows - 1) * (n_cols - 1), 3)
        select : array_like
            Array to select data in the correct order with the same shape as
            the index array.
        """
        # Build the default grid of index :
        nr, nc = self.nrows, self.ncols
        g = np.arange(nr * nc).reshape(nc, nr)
        g = np.fliplr(g.T)  # np.fliplr(np.flipud(g.T))
        # Build indices for one map :
        index = np.zeros((2 * (nr - 1) * (nc - 1), 3))
        q = 0
        for k in range(nc - 1, 0, -1):
            for i in range(nr - 1, 0, -1):
                index[q, :] = [g[i, k], g[i - 1, k], g[i, k - 1]]
                index[q + 1, :] = [g[i - 1, k], g[i, k - 1], g[i - 1, k - 1]]
                q += 2
        # Repeat indices for each map :
        select = np.zeros((self.n, nr, nc), dtype=int)
        l = index.shape[0]  # noqa
        idx = np.zeros((self.n * l, 3), dtype=np.uint32)
        for k in range(self.n):
            idx[k * l:(k + 1) * l, :] = index + k * (nr * nc)
            select[k, ...] = g + k * (nr * nc)
        return idx, select

    def set_data(self, width=None, height=None, dxyz=None, **kwargs):
        """Convert data into a compatible colormap.

        Parameters
        ----------
        kwargs : dict | {}
            Optional args passed to the array2colormap function.

        Returns
        -------
        cmap : array_like
            The colormap of shape (n_sources, n_rows, n_cols, RGBA).
        """
        # Update width/heigth :
        needupdate = False
        if width is not None:
            self.w = width
            needupdate = True
        if height is not None:
            self.h = height
            needupdate = True
        if dxyz is not None:
            self._dxyz = np.array(dxyz)
            needupdate = True
        if needupdate:
            a_position = self._data_to_pos(self._pos)
            self._pos_buffer.set_data(a_position)
        # Update color properties :
        _limits = (self._data.min(), self._data.max())
        cmap = cmap_to_glsl(limits=_limits, **kwargs)
        self.shared_program['texture2D_LUT'] = cmap.texture_lut()
        fcn_color = visuals.shaders.Function(cmap.glsl_map)
        self.shared_program.vert['color'] = fcn_color('v_data')
        self.update()

    # ----------- ALPHA -----------
    @property
    def alpha(self):
        """Get the alpha value."""
        return self._alpha

    @alpha.setter
    def alpha(self, value):
        """Set alpha value."""
        self.shared_program.frag['u_alpha'] = value
        self._alpha = value
        self.update()


# Auto-generate a Visual+Node class for use in the scenegraph.
PicMesh = scene.visuals.create_visual_node(PicVisual)



"""Visual objects of sleep module.

This file contains and initialize visual objects (channel plot, spectrogram,
hypnogram, indicator, shortcuts)
"""
import numpy as np
import scipy.signal as scpsig
import itertools
import logging

from vispy import scene
import vispy.visuals.transforms as vist
from vispy.scene.visuals import Polygon
from vispy.color import Color
from vispy.visuals.filters import Alpha

from .marker import Markers
from sleep_scoring.utils import (color2vb, PrepareData, cmap_to_glsl)
from sleep_scoring.utils.sleep.event import _index_to_events
from sleep_scoring.visuals import TopoMesh, TFmapsMesh
from sleep_scoring.config import PROFILER

logger = logging.getLogger('visbrain')

__all__ = ("Visuals")


"""
###############################################################################
# OBJECTS
###############################################################################
Classes below are used to create visual objects, display on sevral canvas :
- ChannelPlot : plot data (one signal per channel)
- Spectrogram : on a specific channel
- Hypnogram : sleep stages (can be optional)
"""


class Detection(object):
    """Create a detection object."""

    def __init__(self, channels, time, spincol=None, remcol=None,
                 kccol=None, swcol=None, peakcol=None, mtcol=None,
                 spinsym=None, remsym=None, kcsym=None, swsym=None,
                 peaksym=None, mtsym=None, parent=None, parent_hyp=None):
        """Init."""
        self.items = ['Spindles', 'REM', 'K-complexes', 'Slow waves', 'Peaks',
                      'Muscle twitches']
        self.chans = channels
        self.dict = {}
        self.line = {}
        self.peaks = {}
        self.seg = {}
        col = {'Spindles': spincol, 'REM': remcol, 'K-complexes': kccol,
               'Slow waves': swcol, 'Peaks': peakcol, 'Muscle twitches': mtcol}
        sym = {'Spindles': spinsym, 'REM': remsym, 'K-complexes': kcsym,
               'Slow waves': swsym, 'Peaks': peaksym, 'Muscle twitches': mtsym}
        self.time = time
        self.hyp = Markers(parent=parent_hyp)
        self.hyp.set_gl_state('translucent')
        for num, k in enumerate(self):
            self[k] = {'index': np.array([]), 'color': col[k[1]],
                       'connect': np.array([]), 'sym': sym[k[1]]}
            par = parent[self.chans.index(k[0])]
            if k[1] != 'Peaks':
                self.line[k] = scene.visuals.Line(method='gl', parent=par,
                                                  color=col[k[1]])
                self.line[k].set_gl_state('translucent')
            else:
                pos = np.full((1, 3), -10., dtype=np.float32)
                self.peaks[k] = Markers(pos=pos, parent=par,
                                        face_color=col[k[1]])
                self.peaks[k].set_gl_state('translucent')

    def __iter__(self):
        it = itertools.product(self.chans, self.items)
        for k in it:
            yield k

    def __bool__(self):
        return any([bool(self[k]['index'].size) for k in self])

    def __setitem__(self, key, value):
        self.dict[key] = value

    def __getitem__(self, key):
        return self.dict[key]

    def build_line(self, data):
        """Build detections reports.

        Parameters
        ----------
        data : array_like
            Data vector for a spcefic channel.
        """
        for num, k in enumerate(self):
            if self[k]['index'].size:
                # Get the channel number :
                nb = self.chans.index(k[0])
                # Send data :
                if k[1] == 'Peaks':
                    # Get index and channel number :
                    index = self[k]['index'][:, 0]
                    z = np.full(len(index), 2., dtype=np.float32)
                    pos = np.vstack((self.time[index], data[nb, index], z)).T
                    self.peaks[k].set_data(pos=pos, edge_width=0.,
                                           face_color=self[k]['color'])
                else:
                    # Get index and channel number :
                    index = _index_to_events(self[k]['index'])
                    z = np.full(index.shape, 2., dtype=np.float32)
                    # Build position vector :
                    pos = np.vstack((self.time[index], data[nb, index], z)).T
                    # Build connections :
                    connect = np.gradient(index) == 1.
                    connect[0], connect[-1] = True, False
                    self.line[k].set_data(pos=pos, width=4., connect=connect)

    def build_hyp(self, chan, types):
        """Build hypnogram report.

        Parameters
        ----------
        chan : str
            String name of the channel.
        types : str
            String name of the detection type.
        """
        # Get index :
        index = self[(chan, types)]['index']
        # Get only starting points :
        start = index[:, 0]
        y = np.full_like(start, 1.5, dtype=float)
        z = np.full_like(start, -2., dtype=float)
        pos = np.vstack((self.time[start], y, z)).T
        # Set hypnogram data :
        self.hyp.set_data(pos=pos, symbol=self[(chan, types)]['sym'],
                          face_color=self[(chan, types)]['color'],
                          edge_width=1.,
                          edge_color=self[(chan, types)]['color'])

    def visible(self, viz, chan, types):
        """Set channel visibility.

        Parameters
        ----------
        viz : bool
            Boolean value indicating if the plot have to be displayed.
        chan : str
            Channel name.
        types : str
            Detection type name.
        """
        self.hyp.visible = viz
        if types == 'Peaks':
            self.peaks[(chan, types)].visible = viz
        else:
            self.line[(chan, types)].visible = viz

    def delete(self, chan, types):
        """Delete data of a channel."""
        # Remove data from dict :
        self[(chan, types)]['index'] = np.array([])
        # Remove data from plot :
        pos = np.full((1, 3), -10., dtype=np.float32)
        if types == 'Peaks':
            self.peaks[(chan, types)].set_data(pos=pos)
        else:
            self.line[(chan, types)].set_data(pos=pos,
                                              connect=np.array([False]))
        # Remove data from hypnogram :
        self.hyp.set_data(pos=pos)

    def nonzero(self):
        """Return the list of channels with non-empty detections."""
        chans = {}
        for k in self.chans:
            types = []
            for i in self.items:
                if self[(k, i)]['index'].size:
                    types.append(i)
            if types:
                chans[k] = types
        return chans

    def update_keys(self, newkeys):
        """Update the keys of dictionaries."""
        # Get old keys :
        oldkeys = list(self.dict.keys())
        # Check that new keys lentgh has the same size as old keys :
        if len(newkeys) != len(self.chans):
            raise ValueError("The length of new keys must be the same as old"
                             "keys")
        for k in oldkeys:
            # Find index of channel :
            idx = self.chans.index(k[0])
            # Build new key :
            nkey = (newkeys[idx], k[1])
            # Update keys (if needed):
            if nkey not in oldkeys:
                # Update dict and line :
                self.dict[nkey] = self.dict[k]
                if k[1] == 'Peaks':
                    self.peaks[nkey] = self.peaks[k]
                    del self.peaks[k]
                else:
                    self.line[nkey] = self.line[k]
                    del self.line[k]
                # Remove old key :
                del self.dict[k]
        self.chans = newkeys

    def reset(self):
        """Reset all detections."""
        for k in self:
            self[k]['index'] = np.array([])


class ChannelPlot(PrepareData):
    """Plot each channel."""

    def __init__(self, channels, time, color=(.2, .2, .2), width=1.5,
                 color_detection='red', method='gl', camera=None,
                 parent=None, fcn=None):
        # Initialize PrepareData :
        PrepareData.__init__(self, axis=1)

        # Variables :
        self._camera = camera
        self._preproc_channel = -1
        self.rect = []
        self.width = width
        self.autoamp = False
        self._fcn = fcn
        self.visible = np.array([True] + [False] * (len(channels) - 1))
        self.consider = np.ones((len(channels),), dtype=bool)
        self.stage_colors_enabled = False

        # Get color :
        self.color = color2vb(color)
        self.color_detection = color2vb(color_detection)

        # Create one line per channel :
        pos = np.zeros((1, 3), dtype=np.float32)
        self.mesh, self.report, self.grid, self.peak, self.stage_underlays = [], [], [], [], []
        self.loc, self.node = [], []

        self.stage_colors = {}
        self.stage_colors[0] = Color("#56bf8b") # wake
        self.stage_colors[1] = Color("#aabcce") # N1
        self.stage_colors[2] = Color("#405c79") # N2
        self.stage_colors[3] = Color("#0b1c2c") # N3
        self.stage_colors[4] = Color("#bf5656") # REM
        self.stage_colors[-1] = Color("#8bbf56") # art

        for i, k in enumerate(channels):
            # ----------------------------------------------
            # Create a node parent :
            node = scene.Node(name=k + 'plot')
            node.parent = parent[i].wc.scene
            self.node.append(node)

            # ----------------------------------------------
            # (port-visbrain) Create list for tracking stage coloring underlays
            self.stage_underlays.append([])

            # ----------------------------------------------
            # Create main line (for channel plot) :
            mesh = scene.visuals.Line(pos, name=k + 'plot', color=self.color,
                                      method=method, parent=node)
            # force line on top of epoch coloring
            mesh.transform = vist.STTransform(translate=(0, 0, -100))
            mesh.set_gl_state('translucent')
            self.mesh.append(mesh)

            # ----------------------------------------------
            # Create marker peaks :
            mark = Markers(pos=np.zeros((1, 3), dtype=np.float32),
                           parent=node)
            mark.set_gl_state('translucent')
            mark.visible = False
            self.peak.append(mark)

            # ----------------------------------------------
            # Locations :
            loc = scene.visuals.Line(pos, name=k + 'location', method=method,
                                     color=(.1, .1, .1, .3), parent=node,
                                     connect='segments')
            loc.set_gl_state('translucent')
            self.loc.append(loc)

            # ----------------------------------------------
            # Create a grid :
            grid = scene.visuals.GridLines(color=(.1, .1, .1, .5),
                                           scale=(1., .1),
                                           parent=parent[i].wc.scene)
            grid.set_gl_state('translucent')
            self.grid.append(grid)

    def __iter__(self):
        """Iterate over visible mesh."""
        for i, k in enumerate(self.mesh):
            if self.visible[i]:
                yield i, k

    def __len__(self):
        """Return the number of channels."""
        return len(self.mesh)

    def set_data(self, sf, data, hypno, time, sl=None, ylim=None, autoamp=True):
        """Set data to channels.

        Parameters
        ----------
        data: array_like
            Array of data of shape (n_channels, n_points)
        time: array_like
            The time vector.
        sl : slice | None
            A slice object for the time selection of data.
        ylim : array_like | None
            Y-limits of each channel. Must be a (n_channels, 2) array.
        """
        if ylim is None:
            ylim = np.array([data.min(1), data.max(1)]).T

        # Manage slice :
        sl = slice(0, data.shape[1]) if sl is None else sl

        # Slice selection (of time and data) :
        time_sl = time[sl]
        self.x = (time_sl.min(), time_sl.max())
        data_sl = data[self.visible, sl]
        # TODO: (port-visbrain)
        hypno_sl = hypno[sl]
        # fill axis with halfway stuff
        z = np.full_like(time_sl, .5, dtype=np.float32)

        # Prepare the data (only if needed) :
        if self:
            if self._preproc_channel == -1:  # prepare all channels
                data_sl = self._prepare_data(sf, data_sl.copy(), time_sl)
            else:  # filt only one channel
                # Get on which visible channel to apply preprocessing :
                chan_lst_viz = list(np.arange(len(self))[self.visible])
                to_chan = chan_lst_viz.index(self._preproc_channel)
                data_sl[[to_chan], :] = self._prepare_data(sf, data_sl[
                    [to_chan], :].copy(), time_sl)

        # Set data to each plot :
        for l, (i, k) in enumerate(self):
            # ________ MAIN DATA ________
            # Select channel ;
            datchan = data_sl[l, :]

            # Concatenate time / data / z axis :
            dat = np.vstack((time_sl, datchan, z)).T

            # ________ CAMERA ________
            # Use either auto / fixed adaptative camera :
            ycam = (datchan.min(), datchan.max()) if self.autoamp else ylim[i]

            # Get camera rectangle and set it:
            rect = (self.x[0], ycam[0], self.x[1] - self.x[0],
                    ycam[1] - ycam[0])
            self._camera[i].rect = rect

            # delete all underlays
            for poly in self.stage_underlays[i]:
                poly.parent = None
            self.stage_underlays[i] = []
            if self.stage_colors_enabled:
                # calculate % of bar x-axis along which change in stage occurs
                change_points_idx = np.copy(np.where(np.diff(hypno_sl) != 0)[0])
                change_points = np.empty(0)
                for pt_i in range(len(change_points_idx)):
                    change_points = np.append(change_points, float(change_points_idx[pt_i]) / float(len(hypno_sl)))

                change_points = np.concatenate([[0],change_points,[1]])
                stages = [hypno_sl[pt_i] for pt_i in change_points_idx]
                # change_points doesn't include the last stage, so get the stage
                # on the index after the last change point to get the color of the last stage
                stages.append(hypno_sl[-1:][0])

                epoch_rect_arr = []

                for pt_i in range(len(change_points)-1):
                    left = self.x[0] + (self.x[1] - self.x[0]) * change_points[pt_i]
                    right = self.x[0] + (self.x[1] - self.x[0]) * change_points[pt_i+1]
                    width = right - left

                    # z-value 100 to force under line
                    epoch_rect = [
                        (left, ycam[0], 100),
                        (right, ycam[0], 100),
                        (right, ycam[1], 100),
                        (left, ycam[1], 100)
                    ]
                    epoch_rect_arr.append(epoch_rect)

                for pt_i in range(len(epoch_rect_arr)):
                    poly = Polygon(epoch_rect_arr[pt_i], color=self.stage_colors[stages[pt_i]], border_color="white",
                                    border_width=3, parent=k.parent)
                    # make transparent
                    poly.attach(Alpha(0.4))
                    self.stage_underlays[i].append(poly)

            # Set main line :
            k.set_data(dat, width=self.width)
            k.update()
            self.rect.append(rect)

    def set_location(self, sf, data, channel, start, end, factor=100.):
        """Set vertical lines for detections."""
        # Get data limits :
        y = (data.min(), data.max())
        # # Build pos :
        pos = np.zeros((4, 3), dtype=np.float32)
        pos[0, 0:2] = [start, y[0]]
        pos[1, 0:2] = [start, y[1]]
        pos[2, 0:2] = [end, y[0]]
        pos[3, 0:2] = [end, y[1]]
        # Set data pos :
        self.loc[channel].set_data(pos=pos, width=2.)

    def clean(self):
        """Clean all the data."""
        # Empty position :
        pos = np.zeros((1, 3), dtype=np.float32)
        for k in range(len(self)):
            # Main mesh :
            self.mesh[k].set_data(pos=pos, color='gray')
            self.mesh[k].parent = None
            # Report :
            self.report[k].set_data(pos=pos, color='gray')
            self.report[k].parent = None
            # Grid :
            self.grid[k].parent = None
            # Peak locations :
            self.peak[k].set_data(pos=pos, face_color='gray')
            self.peak[k].parent = None
            # Vertical lines :
            self.loc[k].set_data(pos=pos, color='gray')
            self.loc[k].parent = None
        self.mesh, self.report, self.grid, self.peak = [], [], [], []
        self.loc = []

    # ----------- PARENT -----------
    @property
    def parent(self):
        """Get the parent value."""
        return self.mesh[0].parent

    @parent.setter
    def parent(self, value):
        """Set parent value."""
        for i, k, in zip(value, self.mesh):
            k.parent = i.wc.scene

    # ----------- AUTOAMP -----------
    @property
    def autoamp(self):
        """Get the autoamp value."""
        return self._autoamp

    @autoamp.setter
    def autoamp(self, value):
        """Set autoamp value."""
        self._autoamp = value


class Spectrogram(PrepareData):
    """Create and manage a Spectrogram object.

    After object creation, use the set_data() method to pass new data, new
    color, new frequency / time range, new settings...
    """

    def __init__(self, camera, parent=None, fcn=None):
        # Initialize PrepareData :
        PrepareData.__init__(self, axis=0)

        # Keep camera :
        self._camera = camera
        self._rect = (0., 0., 0., 0.)
        self._fcn = fcn

        # Time-frequency map
        self.tf = TFmapsMesh(parent=parent)
        # Spectrogram
        self.mesh = scene.visuals.Image(np.zeros((2, 2)), parent=parent,
                                        name='Fourier transform')
        self.mesh.transform = vist.STTransform()

    def set_data(self, sf, data, time, method='Fourier transform',
                 cmap='rainbow', nfft=30., overlap=0., fstart=.5, fend=20.,
                 contrast=.5, interp='nearest', norm=0):
        """Set data to the spectrogram.

        Use this method to change data, colormap, spectrogram settings, the
        starting and ending frequencies.

        Parameters
        ----------
        sf: float
            The sampling frequency.
        data: array_like
            The data to use for the spectrogram. Must be a row vector.
        time: array_like
            The time vector.
        method: string | 'Fourier transform'
            Computation method.
        cmap : string | 'viridis'
            The matplotlib colormap to use.
        nfft : float | 30.
            Number of fft points for the spectrogram (in seconds).
        overlap : float | .5
            Ovelap proprotion (0 <= overlap <1).
        fstart : float | .5
            Frequency from which the spectrogram have to start.
        fend : float | 20.
            Frequency from which the spectrogram have to finish.
        contrast : float | .5
            Contrast of the colormap.
        interp : string | 'nearest'
            Interpolation method.
        norm : int | 0
            Normalization method for TF.
        """
        # =================== PREPARE DATA ===================
        # Prepare data (only if needed)
        if self:
            data = self._prepare_data(sf, data.copy(), time)

        nperseg = int(round(nfft * sf))

        # =================== TF // SPECTRO ===================
        if method == 'Wavelet':
            self.tf.set_data(data, sf, f_min=fstart, f_max=fend, cmap=cmap,
                             contrast=contrast, n_window=nperseg,
                             overlap=overlap, window='hamming', norm=norm)
            self.tf._image.interpolation = interp
            self.rect = self.tf.rect
            self.freq = self.tf.freqs
        else:
            # =================== CONVERSION ===================
            overlap = int(round(overlap * nperseg))

            if method == 'Multitaper':
                from lspopt import spectrogram_lspopt
                freq, _, mesh = spectrogram_lspopt(data, fs=sf,
                                                   nperseg=nperseg,
                                                   c_parameter=20,
                                                   noverlap=overlap)
            elif method == 'Fourier transform':
                freq, _, mesh = scpsig.spectrogram(data, fs=sf,
                                                   nperseg=nperseg,
                                                   noverlap=overlap,
                                                   window='hamming')
            mesh = 20 * np.log10(mesh)

            # =================== FREQUENCY SELECTION ===================
            # Find where freq is [fstart, fend] :
            f = [0., 0.]
            f[0] = np.abs(freq - fstart).argmin() if fstart else 0
            f[1] = np.abs(freq - fend).argmin() if fend else len(freq)
            # Build slicing and select frequency vector :
            sls = slice(f[0], f[1] + 1)
            freq = freq[sls]
            self._fstart, self._fend = freq[0], freq[-1]

            # =================== COLOR ===================
            # Get clim :
            _mesh = mesh[sls, :]
            contrast = 1. if contrast is None else contrast
            clim = (contrast * _mesh.min(), contrast * _mesh.max())
            # Turn mesh into color array for selected frequencies:
            self.mesh.set_data(_mesh)
            _min, _max = _mesh.min(), _mesh.max()
            _cmap = cmap_to_glsl(limits=(_min, _max), clim=clim, cmap=cmap)
            self.mesh.cmap = _cmap
            self.mesh.clim = 'auto'
            self.mesh.interpolation = interp

            # =================== TRANSFORM ===================
            tm, th = time.min(), time.max()
            # Re-scale the mesh for fitting in time / frequency :
            fact = (freq.max() - freq.min()) / len(freq)
            sc = (th / mesh.shape[1], fact, 1)
            tr = [0., freq.min(), 0.]
            self.mesh.transform.translate = tr
            self.mesh.transform.scale = sc
            # Update object :
            self.mesh.update()
            # Get camera rectangle :
            self.rect = (tm, freq.min(), th - tm, freq.max() - freq.min())
            self.freq = freq
        # Visibility :
        self.mesh.visible = 0 if method == 'Wavelet' else 1
        self.tf.visible = 1 if method == 'Wavelet' else 0

    def clean(self):
        """Clean indicators."""
        pos = np.zeros((3, 4), dtype=np.float32)
        self.mesh.set_data(pos)
        self.mesh.parent = None
        self.mesh = None

    # ----------- RECT -----------
    @property
    def rect(self):
        """Get the rect value."""
        return self._rect

    @rect.setter
    def rect(self, value):
        """Set rect value."""
        self._rect = value
        self._camera.rect = value

    # ----------- INTERP -----------
    @property
    def interp(self):
        """Get the interp value."""
        return self._interp

    @interp.setter
    def interp(self, value):
        """Set interp value."""
        self._interp = value
        self.mesh.interpolation = value
        self.mesh.update()
        self.tf.interpolation = value
        self.tf.update()


class Hypnogram(object):
    """Create a hypnogram object."""

    def __init__(self, time, camera, color='#292824', width=2., parent=None,
                 hconv=None):
        # Keep camera :
        self._camera = camera
        self._rect = (0., 0., 0., 0.)
        self.rect = (time.min(), -5., time.max() - time.min(), 7.)
        self.width = width
        self.n = len(time)
        self._hconv = hconv
        self._hconvinv = {v: k for k, v in self._hconv.items()}
        # Get color :
        self.color = {k: color2vb(color=i) for k, i in zip(color.keys(),
                                                           color.values())}
        # Create a default line :
        pos = np.array([[0, 0], [0, 100]])
        self.mesh = scene.visuals.Line(pos, name='hypnogram', method='gl',
                                       parent=parent, width=width)
        self.mesh._width = width
        self.mesh.set_gl_state('translucent')
        # Create a default marker (for edition):
        self.edit = Markers(parent=parent)
        # self.mesh.set_gl_state('translucent', depth_test=True)
        self.edit.set_gl_state('translucent')
        # Add grid :
        self.grid = scene.visuals.GridLines(color=(.7, .7, .7, 1.),
                                            scale=(30. * time[-1] / len(time),
                                                   1.),
                                            parent=parent)
        self.grid.set_gl_state('translucent')

    def __len__(self):
        """Return the time length."""
        return self.n

    # -------------------------------------------------------------------------
    # SETTING METHODS
    # -------------------------------------------------------------------------
    def set_data(self, sf, data, time, convert=True):
        """Set data to the hypnogram.

        Parameters
        ----------
        sf: float
            The sampling frequency.
        data: array_like
            The data to send. Must be a row vector.
        time: array_like
            The time vector
        convert : bool | True
            Specify if hypnogram data have to be converted.
        """
        # Hypno conversion :
        if (self._hconv != self._hconvinv) and convert:
            data = self.hyp_to_gui(data)
        # Build color array :
        color = np.zeros((len(data), 4), dtype=np.float32)
        for k, v in zip(self.color.keys(), self.color.values()):
            # Set the stage color :
            color[data == k, :] = v
        # Avoid gradient color :
        color[1::, :] = color[0:-1, :]
        # Set data to the mesh :
        self.mesh.set_data(pos=np.vstack((time, -data)).T, width=self.width,
                           color=color)
        self.mesh.update()

    def set_stage(self, stfrom, stend, stage):
        """Add a stage in a specific interval.

        This method only set the stage without updating the entire
        hypnogram.

        Parameters
        ----------
        stfrom : int
            The index where the stage start.
        stend : int
            The index where the stage end.
        stage : int
            Stage value.
        """
        # Convert the stage :
        stagec = self._hconv[stage]
        # Update color :
        self.mesh.color[stfrom + 1:stend + 1, :] = self.color[stagec]
        # Only update the needed part :
        self.mesh.pos[stfrom:stend, 1] = -float(stagec)
        #(port-visbrain) hacky way of forcing mesh to update
        self.mesh.set_data(pos=self.mesh.pos, width=self.width,
                           color=self.mesh.color)
        self.mesh.update()

    def set_grid(self, time, length=30., y=1.):
        """Set grid lentgh."""
        # Get scaling factor :
        sc = (length * time[-1] / len(time), y)
        # Set to the grid :
        self.grid._grid_color_fn['scale'].value = sc
        self.grid.update()

    # -------------------------------------------------------------------------
    # CONVERSION METHODS
    # -------------------------------------------------------------------------
    def hyp_to_gui(self, data):
        """Convert hypnogram data to the GUI.

        Parameters
        ----------
        data : array_like
            The data to send. Must be a row vector.

        Returns
        -------
        datac : array_like
            Converted data
        """
        # Backup copy :
        datac = data.copy()
        data = np.zeros_like(datac)
        # Fill new data :
        for k in self._hconv.keys():
            data[datac == k] = self._hconv[k]
        return data

    def gui_to_hyp(self):
        """Convert GUI hypnogram into data.

        Returns
        -------
        data : array_like
            The converted data.
        """
        # Get latest data version :
        datac = -self.mesh.pos[:, 1]
        data = np.zeros_like(datac)
        # Fill new data :
        for k in self._hconvinv.keys():
            data[datac == k] = self._hconvinv[k]
        return data

    def clean(self, sf, time):
        """Clean indicators."""
        # Mesh :
        posmesh = np.zeros((len(self),), dtype=np.float32)
        self.set_data(sf, posmesh, time)
        # Edit :
        posedit = np.full((1, 3), -10., dtype=np.float32)
        self.edit.set_data(pos=posedit, face_color='gray')

    # ----------- RECT -----------
    @property
    def rect(self):
        """Get the rect value."""
        return self._rect

    @rect.setter
    def rect(self, value):
        """Set rect value."""
        self._rect = value
        self._camera.rect = value


"""
###############################################################################
# TOPOPLOT
###############################################################################
Topoplot class that inherit from the visual TopoMesh and PrepareData for
filetring, de-meaning...
"""


class TopoSleep(TopoMesh, PrepareData):
    """Topoplot for sleep data."""

    def __init__(self, **kwargs):
        # Initialize TopoMesh and PrepareData :
        TopoMesh.__init__(self, **kwargs)
        PrepareData.__init__(self, axis=1)
        # Initialize data, clim, cmap and cblabel :
        self._data = None
        self._clim = None
        self._cmap = None
        self._cblabel = None

    def set_sleep_topo(self, data=None, clim=None, cmap=None, cblabel=None):
        """Send data to TopoGraphic plot."""
        # Data :
        if data is None:
            data = self._data
        self._data = data
        # Clim :
        if clim is None:
            clim = self._clim
        self._clim = clim
        # Cmap :
        if cmap is None:
            cmap = self._cmap
        self._cmap = cmap
        # Cblabel :
        if cblabel is None:
            cblabel = self._cblabel
        self._cblabel = cblabel

        if data is not None:
            self.set_data(data, cmap=cmap, cblabel=cblabel, clim=clim)


"""
###############################################################################
# INDICATORS
###############################################################################
Visual indicators can be used to help the user to see in which time window the
signal is currently plotted. Those indicators are two vertical lines displayed
on the spectrogram and hypnogram.
"""


class Indicator(object):
    """Create a visual indicator (for spectrogram and hypnogram)."""

    def __init__(self, name='indicator', alpha=.3, visible=True, parent=None):
        # Create a vispy image object :
        image = color2vb('gray', alpha=alpha)[np.newaxis, ...]
        self.mesh = scene.visuals.Image(data=image, name=name,
                                        parent=parent)
        self.mesh.transform = vist.STTransform()
        self.mesh.visible = visible

    def set_data(self, xlim, ylim):
        """Move the visual indicator.

        Parameters
        ----------
        xlim : tuple
            A tuple of two float indicating where xlim start and xlim end.
        ylim : tuple
            A tuple of two floats indicating where ylim start and ylim end.
        """
        tox = (xlim[0], ylim[0], -1.)
        sc = (xlim[1] - xlim[0], ylim[1] - ylim[0], 1.)
        # Move the square
        self.mesh.transform.translate = tox
        self.mesh.transform.scale = sc

    def clean(self):
        """Clean indicators."""
        self.mesh.parent = None
        self.mesh = None


"""
###############################################################################
# SHORTCUTS
###############################################################################
Shortcuts applied on each canvas.
"""


class CanvasShortcuts(object):
    """This class add some shortcuts to the main canvas.

    It's also use to initialize to panel of shortcuts.

    Parameters
    ----------
    canvas : vispy canvas
        Vispy canvas to add the shortcuts.
    """

    def __init__(self, canvas):
        """Init."""
        self.sh = [('n', 'Go to the next window'),
                   ('b', 'Go to the previous window'),
                   ('-', 'Decrease amplitude'),
                   ('+', 'Increase amplitude'),
                   ('s', 'Display / hide spectrogram'),
                   ('t', 'Display / hide topoplot'),
                   ('h', 'Display / hide hypnogram'),
                   ('p', 'Display / hide navigation bar'),
                   ('x', 'Display / hide time axis'),
                   ('g', 'Display / hide time grid'),
                   ('z', 'Enable / disable zooming'),
                   ('i', 'Enable / disable indicators'),
                   ('a', 'Scoring: set current window to Art (-1)'),
                   ('w', 'Scoring: set current window to Wake (0)'),
                   ('1', 'Scoring: set current window to N1 (1)'),
                   ('2', 'Scoring: set current window to N2 (2)'),
                   ('3', 'Scoring: set current window to N3 (3)'),
                   ('r', 'Scoring: set current window to REM (4)'),
                   ('Double clik', 'Insert annotation'),
                   ('CTRL + left click', 'Magnify signal under the cursor'),
                   ('CTRL + Num', 'Display the channel Num'),
                   ('CTRL + s', 'Save hypnogram'),
                   ('CTRL + t', 'Display shortcuts'),
                   ('CTRL + e', 'Display documentation'),
                   ('CTRL + d', 'Display / hide setting panel'),
                   ('CTRL + n', 'Take a screenshot'),
                   ('CTRL + q', 'Close Sleep graphical interface'),
                   ]

        # Add shortcuts to vbCanvas :
        @canvas.events.key_press.connect
        def on_key_press(event):
            """Executed function when a key is pressed on a keyboard over canvas.

            :event: the trigger event
            """
            if event.text == ' ':
                pass

            # ------------ SLIDER ------------
            elif event.text.lower() == 'n':  # Next (slider)
                self._SlGoto.setValue(
                    self._SlGoto.value() + self._SigSlStep.value())
            elif event.text.lower() == 'b':  # Before (slider)
                self._SlGoto.setValue(
                    self._SlGoto.value() - self._SigSlStep.value())

            # ------------ AMPLITUDE ------------
            elif event.text in ['-', '+']:  # Decrease / increase amplitude
                delta = 2.5
                sign = 2 * ['-', '+'].index(event.text) - 1
                if self._PanAmpSym.isChecked():  # Symetric amplitudes :
                    val_sym = self._PanAllAmpMax.value() - 2 * sign * delta
                    self._PanAllAmpMax.setValue(val_sym)
                else:  # non-symetrical amplitudes
                    for m, M in zip(self._yminSpin, self._ymaxSpin):
                        m.setValue(m.value() + sign * delta)
                        M.setValue(M.value() - sign * delta)

            # ------------  GRID/MAGNIFY ------------
            elif event.text.lower() == 'm':  # Magnify
                viz = self._slMagnify.isChecked()
                self._slMagnify.setChecked(not viz)
                self._fcn_slider_magnify()

            elif event.text.lower() == 'g':  # Grid
                viz = self._slGrid.isChecked()
                self._slGrid.setChecked(not viz)
                self._fcn_grid_toggle()

            # ------------ SCORING ------------
            elif event.text.lower() == 'a':  # Art
                self._add_stage_on_win(-1)
                self._SlGoto.setValue(self._SlGoto.value(
                ) + self._SigSlStep.value())
                logger.info("Art stage inserted")
            elif event.text.lower() == 'w':  # Wake
                self._add_stage_on_win(0)
                self._SlGoto.setValue(self._SlGoto.value(
                ) + self._SigSlStep.value())
                logger.info("Wake stage inserted")
            elif event.text == '1':
                self._add_stage_on_win(1)
                self._SlGoto.setValue(self._SlGoto.value(
                ) + self._SigSlStep.value())
                logger.info("N1 stage inserted")
            elif event.text == '2':
                self._add_stage_on_win(2)
                self._SlGoto.setValue(self._SlGoto.value(
                ) + self._SigSlStep.value())
                logger.info("N2 stage inserted")
            elif event.text == '3':
                self._add_stage_on_win(3)
                self._SlGoto.setValue(self._SlGoto.value(
                ) + self._SigSlStep.value())
                logger.info("N3 stage inserted")
            elif event.text.lower() == 'r':
                self._add_stage_on_win(4)
                self._SlGoto.setValue(self._SlGoto.value(
                ) + self._SigSlStep.value())
                logger.info("REM stage inserted")

        @canvas.events.mouse_release.connect
        def on_mouse_release(event):
            """Executed function when the mouse is pressed over canvas.

            This method set the transformation to the canvas to NullTransform.
            """
            # Get canvas name :
            name = canvas.title
            condition = bool(name.find('Canvas') + 1)
            if condition and not self._slMagnify.isChecked():
                # Get channel name :
                chan = name.split('Canvas_')[1]
                # Get index :
                idx = self._channels.index(chan)
                # Build transformation :
                self._chan.node[idx].transform = vist.NullTransform()

        @canvas.events.mouse_double_click.connect
        def on_mouse_double_click(event):
            """Executed function when double click mouse over canvas.

            :event: the trigger event
            """
            # Get canvas title :
            is_sp_hyp = canvas.title in ['Hypnogram', 'Spectrogram']
            title = canvas.title if is_sp_hyp else canvas.title.split('_')[1]
            # Annotate the timing :
            if is_sp_hyp:
                cursor = self._time[-1] * event.pos[0] / canvas.size[0]
            else:
                val = self._SlVal.value()
                step = self._SigSlStep.value()
                win = self._SigWin.value()
                tm, th = (val * step, val * step + win)
                cursor = tm + ((th - tm) * event.pos[0] / canvas.size[0])
            # Set the current tab to the annotation tab :
            self.QuickSettings.setCurrentIndex(5)
            # Run annotation :
            self._fcn_annotate_add('', (cursor, cursor), title)

        @canvas.events.mouse_move.connect
        def on_mouse_move(event):
            """Executed function when the mouse move over canvas.

            Magnify for all channels under cursor locations.
            """
            # Get mouse cursor position for the specified canvas :
            zoom = self.menuDispZoom.isChecked()
            if canvas.title in ['Hypnogram', 'Spectrogram'] and not zoom:
                cursor = self._time[-1] * event.pos[0] / canvas.size[0]
            else:
                # Get time parameters (window, step, slider value) :
                val = self._SlVal.value()
                step = self._SigSlStep.value()
                win = self._SigWin.value()
                tm, th = (val * step, val * step + win)
                # Convert cursor in time position :
                cursor = tm + ((th - tm) * event.pos[0] / canvas.size[0])
                # Enable/Disable magnify :
                if self._slMagnify.isChecked():
                    for i, k in self._chan:
                        self._chan.node[i].transform.center = (cursor, 0.)
                        k.update()
                    tm, th = self._time.min(), self._time.max()
            # Set time position to the cursor text :
            cursor = np.round(cursor * 1000.) / 1000.
            self._txtCursor.setText('Cursor : ' + str(cursor) + ' sec')

        @canvas.events.mouse_press.connect
        def on_mouse_press(event):
            """Executed function when single click mouse over canvas.

            Magnigy the signal under the mouse cursor only.
            """
            # ------------- MAGNIFY : CTRL + left click -------------
            name = canvas.title
            is_left = self._is_left_click(event)
            is_ctrl = self._is_modifier(event, 'Control')
            condition = bool(name.find('Canvas') + 1) and is_left and is_ctrl
            if condition and not self._slMagnify.isChecked():
                # Get channel name :
                chan = name.split('Canvas_')[1]
                # Get index :
                idx = self._channels.index(chan)
                # Get cursor position :
                val = self._SlVal.value()
                step = self._SigSlStep.value()
                win = self._SigWin.value()
                tm, th = (val * step, val * step + win)
                cursor = tm + ((th - tm) * event.pos[0] / canvas.size[0])
                # Build transformation :
                kwargs = {'center': (cursor, 0.), 'radii': (3, 15), 'mag': 10}
                transform = vist.nonlinear.Magnify1DTransform(**kwargs)
                self._chan.node[idx].transform = transform

        @canvas.events.mouse_wheel.connect
        def on_mouse_wheel(event):
            pass


class Visuals(CanvasShortcuts):
    """Create the visual objects to be added to the scene."""

    def __init__(self):
        """Init."""
        # =================== VARIABLES ===================
        sf, data, time = self._sf, self._data, self._time
        channels, hypno, cameras = self._channels, self._hypno, self._allCams

        # =================== CHANNELS ===================
        self._chan = ChannelPlot(channels, time, camera=cameras[0],
                                 color=self._chancolor, width=self._lw,
                                 color_detection=self._indicol,
                                 parent=self._chanCanvas,
                                 fcn=self._fcn_slider_move)
        PROFILER('Channels', level=1)

        # =================== SPECTROGRAM ===================
        # Create a spectrogram object :
        self._spec = Spectrogram(camera=cameras[1],
                                 fcn=self._fcn_spec_set_data,
                                 parent=self._specCanvas.wc.scene)
        self._spec.set_data(sf, data[0, ...], time, cmap=self._defcmap)
        PROFILER('Spectrogram', level=1)
        # Create a visual indicator for spectrogram :
        self._specInd = Indicator(name='spectro_indic', visible=True, alpha=.3,
                                  parent=self._specCanvas.wc.scene)
        self._specInd.set_data(xlim=(0, 30), ylim=(0, 20))
        PROFILER('Spectrogram indicator', level=1)

        # =================== HYPNOGRAM ===================
        # Create a hypnogram object :
        self._hyp = Hypnogram(time, camera=cameras[2], color=self._hypcolor,
                              width=self._lwhyp, hconv=self._hconv,
                              parent=self._hypCanvas.wc.scene)
        self._hyp.set_data(sf, hypno, time)
        PROFILER('Hypnogram', level=1)
        # Create a visual indicator for hypnogram :
        self._hypInd = Indicator(name='hypno_indic', visible=True, alpha=.3,
                                 parent=self._hypCanvas.wc.scene)
        self._hypInd.set_data(xlim=(0., 30.), ylim=(-6., 2.))
        PROFILER('Hypnogram indicator', level=1)

        # =================== DETECTIONS ===================
        self._detect = Detection(self._channels.copy(), self._time,
                                 self._defspin, self._defrem, self._defkc,
                                 self._defsw, self._defpeaks, self._defmt,
                                 self._spinsym, self._remsym, self._kcsym,
                                 self._swsym, self._peaksym, self._mtsym,
                                 self._chan.node, self._hypCanvas.wc.scene)
        PROFILER('Detections', level=1)

        # =================== TOPOPLOT ===================
        self._topo = TopoSleep(channels=self._channels, margin=.2,
                               parent=self._topoCanvas.wc.scene)
        # Set camera properties :
        cameras[3].rect = self._topo.rect
        cameras[3].aspect = 1.
        self._pan_pick.model().item(3).setEnabled(any(self._topo._keeponly))
        PROFILER('Topoplot', level=1)

        # =================== SHORTCUTS ===================
        vbcanvas = self._chanCanvas + [self._specCanvas, self._hypCanvas]
        for k in vbcanvas:
            CanvasShortcuts.__init__(self, k.canvas)
        self._shpopup.set_shortcuts(self.sh)
        PROFILER('Shortcuts', level=1)



"""Screenshot window and related functions."""
from sleep_scoring.io import write_fig_pyqt, write_fig_canvas, dialog_save
from sleep_scoring.utils import ScreenshotPopup


class UiScreenshot(object):
    """Initialize the screenshot GUI and functions to apply it."""

    def __init__(self):
        """Init."""
        # Get the list of canvas channel names :
        cn = self._channels + ['spectrogram', 'hypnogram', 'topoplot']
        # Create the screenshot GUI :
        self._ssGui = ScreenshotPopup(self._fcn_run_screenshot,
                                      canvas_names=cn)
        # In Sleep module, disable background and transparency :
        self._ssGui._ssBgcolor.hide()
        self._ssGui._ssBgcolorChk.hide()
        self._ssGui._ssTransp.hide()

    def show_gui_screenshot(self):
        """Display the GUI screenhot."""
        self._ssGui.show()

    def _fcn_run_screenshot(self):
        """Run the screenshot."""
        # Get filename :
        filename = dialog_save(self, 'Screenshot', 'screenshot', "PNG (*.PNG)"
                               ";;TIFF (*.tiff);;JPG (*.jpg);;"
                               "All files (*.*)")
        # Get screenshot arguments :
        kwargs = self._ssGui.to_kwargs()

        if kwargs['entire']:  # Screenshot of the entire window
            self._ssGui._ss.close()
            write_fig_pyqt(self, filename)
        else:  # Screenshot of selected canvas
            # Get canvas name :
            name = kwargs['canvas']
            # Remove unsed entries :
            del kwargs['entire'], kwargs['canvas']
            if name == 'spectrogram':
                # Force the spectrogram to be displayed :
                self.menuDispSpec.setChecked(True)
                self._disptog_spec()
                # Get canvas
                canvas = self._specCanvas.canvas
            elif name == 'hypnogram':
                # Force the spectrogram to be displayed :
                self.menuDispHypno.setChecked(True)
                self._disptog_hyp()
                # Get canvas
                canvas = self._hypCanvas.canvas
            elif name == 'topoplot':
                # Force the spectrogram to be displayed :
                self.menuDispTopo.setChecked(True)
                self._disptog_topo()
                # Get canvas :
                canvas = self._topoCanvas.canvas
            else:
                # Get the index of this channel :
                kc = self._channels.index(name)
                # Force the channel to be displayed :
                self._chanChecks[kc].setChecked(True)
                self._fcn_chan_viz()
                canvas = self._chanCanvas[kc].canvas
            # Finally, render the canvas :
            write_fig_canvas(filename, canvas=canvas, **kwargs)



from .cameras import *
from .color import *
from .filtering import *
from .gui import *
from .guitools import *
from .logging import *
from .memory import *
from .mesh import *
from .others import *
from .physio import *
from .picture import *
from .sigproc import *
from .sleep import *
from .transform import *
from .wrappers import *



"""Load and save configurations."""

__all__ = ['save_config_json', 'load_config_json']


def save_config_json(filename, config):
    """Save configuration file as JSON.

    Parameters
    ----------
    filename : string
        Name of the configuration file to save.
    config : dict
        Dictionary of arguments to save.
    """
    import io
    import json

    # Ensure Python version compatibility
    try:
        to_unicode = unicode
    except NameError:
        to_unicode = str

    if filename:
        with io.open(filename, 'w', encoding='utf8') as f:
            str_ = json.dumps(config, indent=4, sort_keys=True,
                              separators=(',', ': '),  # Pretty printing
                              ensure_ascii=False)
            f.write(to_unicode(str_))


def load_config_json(filename):
    """Load configuration file as JSON.

    Parameters
    ----------
    filename : string
        Name of the configuration file to load.

    Returns
    -------
    config : dict
        Dictionary of config.
    """
    import json
    with open(filename) as f:
        # Load the configuration file :
        config = json.load(f)
    return config



"""This script contains some custom cameras behavior."""
import logging
import numpy as np

from vispy.scene.cameras import PanZoomCamera, TurntableCamera

__all__ = ['FixedCam', 'rotate_turntable', 'optimal_scale_factor',
           'merge_cameras']

logger = logging.getLogger('visbrain')


class FixedCam(PanZoomCamera):
    """Fixed camera but with all PanZoomCamera controls."""

    def __init__(self, *args, **kwargs):
        """Init."""
        PanZoomCamera.__init__(self, *args, **kwargs)

    def viewbox_mouse_event(self, event):
        """Ignore mouse event."""
        pass


def _default_cam_config(use_for, values):
    for k in use_for:
        FIXED_CAM[k] = values


FIXED_CAM = {}
_default_cam_config(['left', 'sagittal_0', 'sagittal'], (-90., 0., [1, 2], -1))
_default_cam_config(['right', 'sagittal_1'], (90., 0., [1, 2], -1))
_default_cam_config(['front', 'coronal_0', 'coronal'], (180., 0., [0, 2], -1))
_default_cam_config(['back', 'coronal_1'], (0., 0., [0, 2], -1))
_default_cam_config(['top', 'axial_0', 'axial'], (0., 90., [0, 1], 1))
_default_cam_config(['bottom', 'axial_1'], (0., -90., [0, 1], 1))


def rotate_turntable(fixed=None, camera_state={}, camera=None,
                     xyz=None, csize=None, margin=1., _scale=1.):
    """Rotate a scene that contains a turntable camera.

    Parameters
    ----------
    fixed : {'left', 'right', 'front', 'back', 'top', 'bottom'}
        Fixed rotation.
    camera_state : dict | {}
        Dictionary to pass to the camera.
    camera : vispy.scene.cameras.TurntableCamera | None
        The camera to update.
    xyz : array_like | None
        The size of the object along the x, y and z axis.
    csize : tuple | None
        The canvas size.
    margin : float | 1.
        Add magin to the scale_factor.
    _scale : float | 1.
        Specifie if the scale_factor have to be rescaled.

    Returns
    -------
    camera_state : dict
        The camera state used.
    """
    assert all([isinstance(k, (int, float)) for k in [margin, _scale]])
    if isinstance(fixed, str):
        azimuth, elevation, idx, priority = FIXED_CAM[fixed]
        camera_state['azimuth'] = azimuth
        camera_state['elevation'] = elevation
    if (xyz is not None) and (len(xyz) == 3):
        xyz = np.asarray(xyz)
        if csize is None:
            camera_state['scale_factor'] = xyz[priority]
        else:
            logger.debug("Optimal scale factor using %r axis" % idx)
            assert len(csize) == 2
            sc = xyz[idx]
            camera_state['scale_factor'] = optimal_scale_factor(sc, csize)
    if 'scale_factor' in camera_state:
        camera_state['scale_factor'] *= _scale * margin
    if camera_state and isinstance(camera, TurntableCamera):
        if 'distance' in camera_state.keys():
            dist = camera_state.pop('distance')
            camera.distance = dist
        camera.set_state(**camera_state)
    return camera_state


def optimal_scale_factor(axis_scale, csize):
    """Get optimal scale_factor according to screen size.

    Parameters
    ----------
    axis_scale : array_like
        The scale along the two axis.
    csize : tuple
        The canvas size.
    """
    assert len(axis_scale) == len(csize)
    x_ratio = axis_scale[0] / csize[0]
    y_ratio = axis_scale[1] / csize[1]
    # Get the optimal scaling factor :
    opt_scale_factor = axis_scale[np.argmax([x_ratio, y_ratio])]
    return opt_scale_factor


def merge_cameras(*args):
    """Merge several TurnTable cameras.

    This function returns the mean center and the max across scale_factor and
    distances.

    Parameter
    ---------
    args : TurnTableCamera
        The turntable cameras to use.

    Returns
    -------
    cam_state : dict
        The dictionary of merged camera states.
    """
    assert all([isinstance(k, TurntableCamera) for k in args])
    center = np.zeros((3,), dtype=np.float32)
    scale_factor, distance = [], []
    # Get the camera state of multiple cameras :
    for k in args:
        cam_state = k.get_state()
        center += cam_state['center']
        scale_factor.append(cam_state['scale_factor'])
        if isinstance(k.distance, (int, float)):
            distance.append(k.distance)
    # Get mean center, max scale_factor and max distance :
    center /= len(args)
    scale_factor = np.max(scale_factor)
    distance = np.max(distance) if len(distance) else None
    # Create turntable camera :
    cam = TurntableCamera(center=center, scale_factor=scale_factor,
                          distance=distance)
    return cam



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages

__version__ = "0.0.1"
NAME = 'sleep_scoring'
AUTHOR = "Visbrain developers"
MAINTAINER = "Anjo Pagdanganan"
EMAIL = 'anjopag31@berkeley.edu'
KEYWORDS = "brain MNI GPU visualization data OpenGL vispy neuroscience " + \
           "sleep data-mining"
DESCRIPTION = "Standalone version of Sleep module from Visbrain suite."
DOWNLOAD_URL = "https://github.com/x64-bit/sleep-scoring"
# Data path :
HERE = os.path.abspath(os.path.dirname(__file__))
PACKAGE_DATA = {}


def read(fname):
    """Read README and LICENSE."""
    return open(os.path.join(HERE, fname), 'rb').read().decode('utf8')


setup(
    # DESCRIPTION
    name=NAME,
    version=__version__,
    description=DESCRIPTION,
    long_description=read('README.rst'),
    keywords=KEYWORDS,
    license=read('LICENSE'),
    author=AUTHOR,
    maintainer=MAINTAINER,
    author_email=EMAIL,
    download_url=DOWNLOAD_URL,
    # PACKAGE / DATA
    packages=find_packages(),
    package_dir={'visbrain': 'visbrain'},
    package_data=PACKAGE_DATA,
    include_package_data=True,
    platforms='any',
    setup_requires=['numpy'],
    install_requires=[
        "numpy>=1.13",
        "scipy",
        "vispy>=0.5.2",
        "matplotlib>=1.5.5",
        "pyqt5",
        "pillow",
        "PyOpenGL"
    ],
    extras_require={
        'full': ["mne", "tensorpac", "pandas", "xlrd", "scikit-image",
                 "nibabel"],
        'sleep': ["mne", "tensorpac"],
        'roi': ["pandas", "xlrd"],
        'topo': ["scikit-image"]
    },
    dependency_links=[],
    classifiers=["Development Status :: 3 - Alpha",
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Education',
                 'Intended Audience :: Developers',
                 'Topic :: Scientific/Engineering :: Visualization',
                #  "Programming Language :: Python :: 3.5",
                #  "Programming Language :: Python :: 3.6",
                #  "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.9",
                 "Programming Language :: Python :: 3.10",
                 "Programming Language :: Python :: 3.11",
                 "Programming Language :: Python :: 3.12",
                 "Operating System :: MacOS",
                #  "Operating System :: POSIX :: Linux",
                 "Operating System :: Microsoft :: Windows",
                 "Natural Language :: English"
                 ])



"""Set of tools to filter data."""

import numpy as np
from scipy.signal import butter, filtfilt, lfilter, bessel, welch, detrend

__all__ = ('filt', 'morlet', 'ndmorlet', 'morlet_power', 'welch_power',
           'PrepareData')

#############################################################################
# FILTERING
#############################################################################


def filt(sf, f, x, btype='bandpass', order=3, method='butterworth',
         way='filtfilt', axis=0):
    """Filt data.

    Parameters
    ----------
    sf : float
        The sampling frequency
    f : array_like
        Frequency vector (2,)
    x : array_like
        The data to filt.
    btype : {'bandpass', 'bandstop', 'highpass', 'lowpass'}
        If highpass, the first value of f will be used. If lowpass
        the second value of f will be used.
    order : int | 3
        The filter order.
    method : {'butterworth', 'bessel'}
        Filter type to use.
    way : {'filtfilt', 'lfilter'}
        Specify if the filter has to be one way ('lfilter') or two ways
        ('filtfilt').
    axis : int | 0
        The axis along which the filter is applied.

    Returns
    -------
    xfilt : array_like
        Filtered data.
    """
    # Normalize frequency vector according to btype :
    if btype in ['bandpass', 'bandstop']:
        fnorm = np.divide(f, .5 * sf)
    elif btype == 'lowpass':
        fnorm = np.array(f[-1] / (.5 * sf))
    elif btype == 'highpass':
        fnorm = np.array(f[0] / (.5 * sf))

    # Get filter coefficients :
    if method == 'butterworth':
        b, a = butter(order, fnorm, btype=btype)
    elif method == 'bessel':
        b, a = bessel(order, fnorm, btype=btype)

    # Apply filter :
    if way == 'filtfilt':
        return filtfilt(b, a, x, axis=axis)
    elif way == 'lfilter':
        return lfilter(b, a, x, axis=axis)

#############################################################################
# WAVELET
#############################################################################


def _morlet_wlt(sf, f, width=7.0):
    """Get a Morlet's wavelet.

    Parameters
    ----------
    sf : float
        Sampling frequency.
    f : array_like
        Frequency vector of shape (2,).
    width : float | 7.0
        Width of the wavelet.
    wlt: array_like
        Morlet wavelet.
    """
    sf, f, width = float(sf), float(f), float(width)
    dt = 1 / sf
    sf = f / width
    st = 1 / (2 * np.pi * sf)

    # Build morlet wavelet :
    t = np.arange(-width * st / 2, width * st / 2, dt)
    a = 1 / np.sqrt((st * np.sqrt(np.pi)))
    wlt = a * np.exp(-np.square(t) / (2 * np.square(st))) * np.exp(
        1j * 2 * np.pi * f * t)

    return wlt


def morlet(x, sf, f, width=7.0):
    """Complex decomposition of a signal x using the morlet wavelet.

    Parameters
    ----------
    x : array_like
        The signal to use for the complex decomposition. Must be
        a vector of length N.
    sf : float
        Sampling frequency
    f : array_like, shape (2,)
        Frequency vector
    width : float | 7.0
        Width of the wavelet

    Returns
    -------
    xout: array_like
        The complex decomposition of the signal x.
    """
    # Get the wavelet :
    m = _morlet_wlt(sf, f, width)

    # Compute morlet :
    y = np.convolve(x, m)
    xout = y[int(np.ceil(len(m) / 2)) - 1:int(len(y) - np.floor(len(m) / 2))]

    return xout


def ndmorlet(x, sf, f, axis=0, get=None, width=7.0):
    """Complex decomposition using Morlet's wlt for a multi-dimentional array.

    Parameters
    ----------
    x : array_like
        The signal to use for the complex decomposition.
    sf : float
        Sampling frequency
    f : array_like
        Frequency vector of shape (2,)
    axis : integer | 0
        Specify the axis where is located the time dimension
    get : {None, 'amplitude', 'phase', 'power'}
        Specify if the amplitude, phase or power of the filtered signal have to
        be returned or only the filtered signal.
    width : float | 7.0
        Width of the wavelet

    Returns
    -------
        xout: array, same shape as x
            Complex decomposition of x.
    """
    # Get the wavelet :
    m = _morlet_wlt(sf, f, width)

    # Define a morlet function :
    def morlet_fcn(xt):
        # Compute morlet :
        y = np.convolve(xt, m)
        return y[int(np.ceil(len(m) / 2)) - 1:int(len(y) - np.floor(
            len(m) / 2))]

    xf = np.apply_along_axis(morlet_fcn, axis, x)
    # Get amplitude / power / phase :
    if get == 'amplitude':
        return np.abs(xf)
    elif get == 'power':
        return np.square(np.abs(xf))
    elif get == 'phase':
        return np.angle(xf)


def morlet_power(x, freqs, sf, norm=True):
    """Compute bandwise-normalized power of data using morlet wavelet.

    Parameters
    ----------
    x : array_like
        Row vector signal.
    freqs : array_like
        Frequency bands for power computation. The power will be computed
        using successive frequency band (e.g freqs=(1., 2, .3)).
    sf : float
        Sampling frequency.
    norm : bool | True
        If True, return bandwise normalized band power
        (For each time point, the sum of power in the 4 band equals 1)

    Returns
    -------
    xpow : array_like
        The power in the specified frequency bands of shape
        (len(freqs)-1, npts).
    """
    # Build frequency vector :
    f = np.c_[freqs[0:-1], freqs[1::]].mean(1)
    # Get wavelet transform :
    # TODO: np.float is deprecated; use builtin float instead
    # xpow = np.zeros((len(f), len(x)), dtype=np.float)
    xpow = np.zeros((len(f), len(x)), dtype=float)
    for num, k in enumerate(f):
        xpow[num, :] = np.abs(morlet(x, sf, k))
    # Compute inplace power :
    np.power(xpow, 2, out=xpow)
    # Normalize by the band sum :
    if norm:
        sum_pow = xpow.sum(0).reshape(1, -1)
        np.divide(xpow, sum_pow, out=xpow)
    return xpow


def welch_power(x, freqs, sf, window_s=10, norm=True):
    """Compute bandwise-normalized power of data using welch power.

    Parameters
    ----------
    x : array_like
        Row vector signal.
    freqs : array_like
        Frequency bands for power computation. The power will be computed
        using successive frequency band (e.g freqs=(1., 2, .3)).
    sf : float
        Sampling frequency.
    window_s : int | 10
        Length of NFFT
    norm : bool | True
        If True, return bandwise normalized band power
        (For each time point, the sum of power in the 4 band equals 1)

    Returns
    -------
    xpow : array_like
        The power in the specified frequency bands of shape
        (len(freqs)-1, npts).
    """
    sf = int(sf)
    freq_spacing = .1
    n_epoch = max(1, int(len(x) / (window_s * sf)))

    # TODO: np.float is deprecated; use builtin float instead
    # xpow = np.zeros((len(freqs) - 1, n_epoch), dtype=np.float)
    xpow = np.zeros((len(freqs) - 1, n_epoch), dtype=float)

    for i in np.arange(0, len(x), window_s * sf):
        f, pxx_spec = welch(x[int(i):int(i + window_s * sf)], sf,
                            nperseg=sf * (1. / freq_spacing),
                            scaling='spectrum')
        epoch = int(i / (window_s * sf))

        for num, k in enumerate(freqs[:-1]):
            fmin = np.abs(f - k).argmin()
            fmax = np.abs(f - freqs[num + 1]).argmin()
            xpow[num, epoch] = np.mean(pxx_spec[fmin:fmax])

    # Normalize by the band sum :
    if norm:
        sum_pow = xpow.sum(0).reshape(1, -1)
        np.divide(xpow, sum_pow, out=xpow)

    # Oversample
    xpow = np.repeat(xpow, int(window_s * sf), axis=1)
    return xpow


class PrepareData(object):
    """Prepare data before plotting.

    This class group a set of signal processing tools including :
        - De-meaning
        - De-trending
        - Filtering
        - Decomposition (filter / amplitude / power / phase)
    """

    def __init__(self, axis=0, demean=False, detrend=False, filt=False,
                 fstart=12., fend=16., forder=3, way='lfilter',
                 filt_meth='butterworth', btype='bandpass', dispas='filter'):
        """Init."""
        # Axis along which to perform preparation :
        self.axis = axis
        # Demean and detrend :
        self.demean = demean
        self.detrend = detrend
        # Filtering :
        self.filt = filt
        self.fstart, self.fend = fstart, fend
        self.forder, self.filt_meth = forder, filt_meth
        self.way, self.btype = way, btype
        self.dispas = dispas

    def __bool__(self):
        """Return if data have to be prepared."""
        return any([self.demean, self.detrend, self.filt])

    def _prepare_data(self, sf, data, time):
        """Prepare data before plotting."""
        # ============= DEMEAN =============
        if self.demean:
            mean = np.mean(data, axis=self.axis, keepdims=True)
            np.subtract(data, mean, out=data)

        # ============= DETREND =============
        if self.detrend:
            data = detrend(data, axis=self.axis)

        # ============= FILTERING =============
        if self.filt:
            if self.dispas == 'filter':
                data = filt(sf, np.array([self.fstart, self.fend]), data,
                            btype=self.btype, order=self.forder, way=self.way,
                            method=self.filt_meth, axis=self.axis)
            else:
                # Compute ndwavelet :
                f = np.array([self.fstart, self.fend]).mean()
                data = ndmorlet(data, sf, f, axis=self.axis, get=self.dispas)

        return data

    def update(self):
        """Update object."""
        if self._fcn is not None:
            self._fcn()



"""Base class for objects that need an iage (TF, Pacmap, spectro...)."""
import numpy as np
import logging

from vispy import scene

from .visbrain_obj import VisbrainObject
from ..utils import vispy_array, wrap_properties, color2vb

logger = logging.getLogger('visbrain')


class ImageObj(VisbrainObject):
    """Create a single image object.

    Parameters
    ----------
    data : array_like
        Array of data. If data.ndim in [1, 2] the color is inferred from
        the data. Otherwise, if data.ndim is 3, data is interpreted as
        color if the last dimension is either 3 (RGB) or 4 (RGBA).
    xaxis : array_like | None
        Vector to use for the x-axis (number of columns in the image).
        If None, `xaxis` is inferred from the second dimension of data.
    yaxis : array_like | None
        Vector to use for the y-axis (number of rows in the image).
        If None, `yaxis` is inferred from the first dimension of data.
    clim : tuple | None
        Colorbar limits. If None, `clim=(data.min(), data.max())`
    cmap : string | None
        Colormap name.
    vmin : float | None
        Minimum threshold of the colorbar.
    under : string/tuple/array_like | None
        Color for values under vmin.
    vmax : float | None
        Maximum threshold of the colorbar.
    under : string/tuple/array_like | None
        Color for values over vmax.
    interpolation : string | 'nearest'
        Interpolation method for the image. See vispy.scene.visuals.Image for
        availables interpolation methods.
    max_pts : int | -1
        Maximum number of points of the image along the x or y axis. This
        parameter is essentially used to solve OpenGL issues with very large
        images.
    transform : VisPy.visuals.transforms | None
        VisPy transformation to set to the parent node.
    parent : VisPy.parent | None
        Markers object parent.
    verbose : string
        Verbosity level.
    kw : dict | {}
        Optional arguments are used to control the colorbar
        (See :class:`ColorbarObj`).

    Notes
    -----
    List of supported shortcuts :

        * **s** : save the figure
        * **<delete>** : reset camera

    Examples
    --------
    >>> import numpy as np
    >>> from sleep_scoring.objects import ImageObj
    >>> n = 100
    >>> time = np.r_[np.arange(n - 1), np.arange(n)[::-1]]
    >>> time = time.reshape(-1, 1) + time.reshape(1, -1)
    >>> im = ImageObj('im', time, cmap='Spectral_r', interpolation='bicubic')
    >>> im.preview(axis=True)
    """

    def __init__(self, name, data=None, xaxis=None, yaxis=None, cmap='viridis',
                 clim=None, vmin=None, under='gray', vmax=None, over='red',
                 interpolation='nearest', max_pts=-1, parent=None,
                 transform=None, verbose=None, **kw):
        """Init."""
        VisbrainObject.__init__(self, name, parent, transform, verbose, **kw)
        self._update_cbar_args(cmap, clim, vmin, vmax, under, over)

        # _______________________ CHECKING _______________________
        assert isinstance(max_pts, int)
        self._n_limit = max_pts  # fix GL issues with large images

        # _______________________ VISUAL _______________________
        self._image = scene.visuals.Image(parent=self._node, name="_ImageObj",
                                          interpolation=interpolation)
        self._image.transform = scene.transforms.STTransform()

        # _______________________ SET DATA _______________________
        if isinstance(data, np.ndarray):
            self.set_data(data, xaxis, yaxis, clim, cmap, vmin, under, vmax,
                          over)

    def set_data(self, data, xaxis=None, yaxis=None, clim=None, cmap=None,
                 vmin=None, under=None, vmax=None, over=None):
        """Set data to the image."""
        assert isinstance(data, np.ndarray) and data.ndim <= 3
        data = np.atleast_2d(data)
        # Get the x-axis and y-axis :
        xaxis = np.arange(data.shape[1]) if xaxis is None else np.array(xaxis)
        yaxis = np.arange(data.shape[0]) if yaxis is None else np.array(yaxis)
        assert (len(xaxis) == data.shape[1]) and (len(yaxis) == data.shape[0])
        self._dim = (xaxis.min(), xaxis.max(), yaxis.min(), yaxis.max())
        # Convert data to color (if data.ndim == 2)
        if data.ndim == 2:  # infer color from data
            clim = (data.min(), data.max()) if clim is None else clim
            # Limit the number of points :
            if any([k > self._n_limit for k in data.shape]):
                dsf_x = max(1, int(np.round(data.shape[0] / self._n_limit)))
                dsf_y = max(1, int(np.round(data.shape[1] / self._n_limit)))
                logger.debug("Image size reduced along the x and y-axis with "
                             "a down-sampling factor of %s" % ([dsf_x, dsf_y]))
                data = data[::dsf_x, ::dsf_y]
                xaxis, yaxis = xaxis[::dsf_x], yaxis[::dsf_y]
            # Set properties :
            self._update_cbar_args(cmap, clim, vmin, vmax, under, over)
            # Get colormap :
            cmap = self._get_glsl_colormap(limits=(data.min(), data.max()))
            self._image.cmap = cmap
            self._image.clim = 'auto'
        else:  # data is already a compatible color
            assert data.shape[-1] in [3, 4]
        # Set color to the image :
        self._image.set_data(vispy_array(data))
        fact_x = (self._dim[1] - self._dim[0]) / len(xaxis)
        fact_y = (self._dim[3] - self._dim[2]) / len(yaxis)
        sc = (fact_x, fact_y, 1.)
        tr = (self._dim[0], self._dim[2], 0.)
        self._image.transform.scale = sc
        self._image.transform.translate = tr

    def _get_camera(self):
        """Get the most adapted camera."""
        rect = (self._dim[0], self._dim[2], self._dim[1] - self._dim[0],
                self._dim[3] - self._dim[2])
        flip = (False, type(self).__name__ == 'ImageObj', False)
        return scene.cameras.PanZoomCamera(rect=rect, flip=flip)

    ###########################################################################
    ###########################################################################
    #                             PROPERTIES
    ###########################################################################
    ###########################################################################

    # ----------- INTERPOLATION -----------
    @property
    def interpolation(self):
        """Get the interpolation value."""
        return self._image.interpolation

    @interpolation.setter
    @wrap_properties
    def interpolation(self, value):
        """Set interpolation value."""
        assert isinstance(value, str)
        self._image.interpolation = value

    # ----------- CMAP -----------
    @property
    def cmap(self):
        """Get the cmap value."""
        return self._cmap

    @cmap.setter
    @wrap_properties
    def cmap(self, value):
        """Set cmap value."""
        self._cmap = value

    # ----------- CLIM -----------
    @property
    def clim(self):
        """Get the clim value."""
        return self._clim

    @clim.setter
    @wrap_properties
    def clim(self, value):
        """Set clim value."""
        assert len(value) == 2
        self._clim = value

    # ----------- VMIN -----------
    @property
    def vmin(self):
        """Get the vmin value."""
        return self._vmin

    @vmin.setter
    @wrap_properties
    def vmin(self, value):
        """Set vmin value."""
        assert isinstance(value, (int, float))
        self._vmin = value

    # ----------- VMAX -----------
    @property
    def vmax(self):
        """Get the vmax value."""
        return self._vmax

    @vmax.setter
    @wrap_properties
    def vmax(self, value):
        """Set vmax value."""
        assert isinstance(value, (int, float))
        self._vmax = value

    # ----------- UNDER -----------
    @property
    def under(self):
        """Get the under value."""
        return self._under

    @under.setter
    @wrap_properties
    def under(self, value):
        """Set under value."""
        color = color2vb(value)
        self._under = color

    # ----------- OVER -----------
    @property
    def over(self):
        """Get the over value."""
        return self._over

    @over.setter
    @wrap_properties
    def over(self, value):
        """Set over value."""
        color = color2vb(value)
        self._over = color



"""Read nifti (nii.gz) files."""
import os

import numpy as np

from .dependencies import is_nibabel_installed
from .path import path_to_visbrain_data
from ..utils.transform import array_to_stt

from vispy.visuals.transforms import (MatrixTransform, ChainTransform,
                                      STTransform)


def read_nifti(path, hdr_as_array=False):
    """Read data from a NIFTI file using Nibabel.

    Parameters
    ----------
    path : string
        Path to the nifti file.

    Returns
    -------
    vol : array_like
        The 3-D volume data.
    header : Nifti1Header
        Nifti header.
    transform : VisPy.transform
        The transformation
    """
    is_nibabel_installed(raise_error=True)
    import nibabel as nib
    # Load the file :
    img = nib.load(path)
    # Get the data and affine transformation ::
    vol = img.get_data()
    affine = img.affine
    # Replace NaNs with 0. :
    vol[np.isnan(vol)] = 0.
    # Define the transformation :
    if hdr_as_array:
        transform = affine
    else:
        transform = array_to_stt(affine)

    return vol, img.header, transform


def read_mist(name):
    """Load MIST parcellation.

    See : MIST: A multi-resolution parcellation of functional networks
    Authors : Sebastian Urchs, Jonathan Armoza, Yassine Benhajali,
              Jolène St-Aubin, Pierre Orban, Pierre Bellec

    Parameters
    ----------
    name : string
        Name of the level. Use MIST_x with x 7, 12, 20, 36, 64, 122 or ROI.

    Returns
    -------
    vol : array_like | None
        ROI volume.
    labels : array_like | None
        Array of labels.
    index : array_like | None
        Array of index that make the correspondance between the volume values
        and labels.
    hdr : array_like | None
        Array of transform source's coordinates into the volume space.
    """
    name = name.upper()
    assert ('MIST' in name) and ('_' in name)
    level = name.split('_')[-1]
    assert level in ['7', '12', '20', '36', '64', '122', 'ROI']
    # Define path :
    parc, parc_info = '%s.nii.gz', '%s.csv'
    folder, folder_info = 'Parcellations', 'Parcel_Information'
    mist_path = path_to_visbrain_data('mist', 'roi')
    parc_path = os.path.join(*(mist_path, folder, parc % name))
    parc_info_path = os.path.join(*(mist_path, folder_info, parc_info % name))
    # Load info :
    m = np.genfromtxt(parc_info_path, delimiter=';', dtype=str, skip_header=1,
                      usecols=[0, 1, 2])
    n_roi = m.shape[0]
    index = m[:, 0].astype(int)
    lab_, name_ = 'label_%s' % level, 'name_%s' % level
    labels = np.zeros(n_roi, dtype=[(lab_, object), (name_, object)])
    labels[lab_] = m[:, 1]
    labels[name_] = np.char.replace(np.char.capitalize(m[:, 2]), '_', ' ')
    # Load parc :
    vol, _, hdr = read_nifti(parc_path, hdr_as_array=True)
    return vol, labels, index, hdr


def _niimg_var(vol, hdr):
    """Get transformation variables.

    Parameters
    ----------
    vol : array_like
        The 3D array of the volume.
    hdr : array_like
        The (4, 4) transformation array.

    Returns
    -------
    sh : array_like
        Shape of the volume
    diag : array_like
        Diagonale of the transformation
    tr : array_like
        Translation of the transformation
    """
    assert vol.ndim == 3, "Volume should be an (n_x, n_y, n_z) array"
    n_x, n_y, n_z = vol.shape
    assert isinstance(hdr, (MatrixTransform, np.ndarray))
    if isinstance(hdr, MatrixTransform):
        affine = np.array(hdr.matrix).copy()
    # Get diagonal and translation
    d_x, d_y, d_z = np.diag(affine)[0:-1]
    t_x, t_y, t_z = affine[-1, 0:-1]
    return np.array(vol.shape), np.diag(affine)[0:-1], affine[-1, 0:-1]


def _niimg_norm(sh, diag, translate):
    """Normalize the volume between (0., 1.)."""
    # Compute normalization ratio
    ratio = np.abs(diag) * sh
    sgn = np.sign(diag)
    # Get scale and translate
    sc = 1. / ratio
    tr = -(translate + np.array([0., 0, 0])) / ratio
    # Define transformations of each slice
    sg_norm = STTransform(scale=(sc[1], sc[2], 1.),
                          translate=(tr[1], tr[2], 1.))
    cr_norm = STTransform(scale=(sc[0], sc[2], 1.),
                          translate=(sgn[0] * tr[0], tr[2], 1.))
    ax_norm = STTransform(scale=(sc[1], sc[0], 1.),
                          translate=(tr[1], sgn[0] * tr[0], 1.))
    return sg_norm, cr_norm, ax_norm


def _niimg_rot():
    """Get rotation trnasformations of each slice."""
    # Sagittal
    sg_rot = MatrixTransform()
    sg_rot.rotate(90., (0, 0, 1))
    sg_rot.rotate(180., (0, 1, 0))
    # Coronal
    cr_rot = MatrixTransform()
    cr_rot.rotate(90., (0, 0, 1))
    cr_rot.rotate(180., (0, 1, 0))
    # Axial
    ax_rot = MatrixTransform()
    ax_rot.rotate(180., (1, 0, 0))
    return sg_rot, cr_rot, ax_rot


def _niimg_mat(hdr, idx):
    """Get the transformation of a single slice.

    Parameters
    ----------
    hdr : array_like
        The (4, 4) transformation array.
    idx : tuple
        Slices indicies.

    Returns
    -------
    tf : MatrixTransform
        Image transformation.
    """
    hdr_mat = np.array(hdr.matrix).copy().T
    mat = np.identity(4, dtype=np.float32)

    to_idx = [[idx[0]], [idx[1]]], [idx[0], idx[1]]
    mat[[[0], [1]], [0, 1]] = hdr_mat[to_idx]
    mat[[0, 1], -1] = hdr_mat[[idx[0], idx[1]], -1]
    return MatrixTransform(mat.T)


def _niimg_mni(hdr):
    """Transformation for MNI conversions of each slice."""
    sg_mni = _niimg_mat(hdr, (2, 1))
    cr_mni = _niimg_mat(hdr, (2, 0))
    ax_mni = _niimg_mat(hdr, (1, 0))
    return sg_mni, cr_mni, ax_mni


def niimg_to_transform(vol, hdr, as_bgd=True, vol_bgd=None, hdr_bgd=None):
    """Get transformations of nii.gz files for cross-sections.

    Parameters
    ----------
    vol : array_like
        3D volume data.
    hdr : array_like
        Array of transformation of shape (4, 4).
    as_bgd : bool | True
        Specify if the volume is a background image or have to be considered as
        an activation image.
    vol_bgd : array_like | None
        Volume data if `as_bgd` is True.
    hdr_bgd : array_like | None
        Transformation array if `as_bgd` is True.

    Returns
    -------
    sg_tf : ChainTransform
        Transformation of sagittal view
    cr_tf : ChainTransform
        Transformation of coronal view
    ax_tf : ChainTransform
        Transformation of axial view
    """
    # Get transformation variables
    sh_img, diag_img, tr_img = _niimg_var(vol, hdr)
    # Get the normalization transformation depending if the volume is an image
    # background or an activation image
    if as_bgd:  # Background image
        sg_norm, cr_norm, ax_norm = _niimg_norm(sh_img - 1, diag_img, tr_img)
    else:       # Activation image
        sh_bgd, diag_bgd, tr_bgd = _niimg_var(vol_bgd, hdr_bgd)
        sg_norm, cr_norm, ax_norm = _niimg_norm(sh_bgd - 1, diag_bgd, tr_bgd)
    # Get MNI and rotation transformations
    sg_mni, cr_mni, ax_mni = _niimg_mni(hdr)
    sg_rot, cr_rot, ax_rot = _niimg_rot()
    # Build the chain of transformation
    sg_tf = ChainTransform([sg_norm, sg_rot, sg_mni])
    cr_tf = ChainTransform([cr_norm, cr_rot, cr_mni])
    ax_tf = ChainTransform([ax_norm, ax_rot, ax_mni])
    return sg_tf, cr_tf, ax_tf



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/etienne/Toolbox/visbrain/visbrain/visuals/cbar/gui/CbarPyQtGui.ui'
#
# Created by: PyQt5 UI code generator 5.9
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(489, 881)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self._cbar_grp = QtWidgets.QGroupBox(Form)
        self._cbar_grp.setCheckable(True)
        self._cbar_grp.setChecked(False)
        self._cbar_grp.setObjectName("_cbar_grp")
        self.verticalLayout_10 = QtWidgets.QVBoxLayout(self._cbar_grp)
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.CbarLayout = QtWidgets.QVBoxLayout()
        self.CbarLayout.setContentsMargins(-1, 0, -1, -1)
        self.CbarLayout.setObjectName("CbarLayout")
        self.groupBox_4 = QtWidgets.QGroupBox(self._cbar_grp)
        self.groupBox_4.setObjectName("groupBox_4")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.groupBox_4)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.gridLayout_9 = QtWidgets.QGridLayout()
        self.gridLayout_9.setObjectName("gridLayout_9")
        self.object = QtWidgets.QComboBox(self.groupBox_4)
        self.object.setObjectName("object")
        self.gridLayout_9.addWidget(self.object, 0, 2, 1, 1)
        self.line_12 = QtWidgets.QFrame(self.groupBox_4)
        self.line_12.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_12.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_12.setObjectName("line_12")
        self.gridLayout_9.addWidget(self.line_12, 0, 1, 1, 1)
        self.label_18 = QtWidgets.QLabel(self.groupBox_4)
        self.label_18.setObjectName("label_18")
        self.gridLayout_9.addWidget(self.label_18, 0, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_9.addItem(spacerItem, 1, 2, 1, 1)
        self.verticalLayout_7.addLayout(self.gridLayout_9)
        self.CbarLayout.addWidget(self.groupBox_4)
        self.groupBox_3 = QtWidgets.QGroupBox(self._cbar_grp)
        self.groupBox_3.setObjectName("groupBox_3")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.groupBox_3)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.gridLayout_7 = QtWidgets.QGridLayout()
        self.gridLayout_7.setObjectName("gridLayout_7")
        self.border = QtWidgets.QCheckBox(self.groupBox_3)
        font = QtGui.QFont()
        font.setItalic(True)
        self.border.setFont(font)
        self.border.setObjectName("border")
        self.gridLayout_7.addWidget(self.border, 4, 0, 1, 1)
        self.line_7 = QtWidgets.QFrame(self.groupBox_3)
        self.line_7.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_7.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_7.setObjectName("line_7")
        self.gridLayout_7.addWidget(self.line_7, 1, 1, 1, 1)
        self.bw = QtWidgets.QDoubleSpinBox(self.groupBox_3)
        self.bw.setDecimals(1)
        self.bw.setMinimum(0.1)
        self.bw.setObjectName("bw")
        self.gridLayout_7.addWidget(self.bw, 4, 2, 1, 1)
        self.width = QtWidgets.QDoubleSpinBox(self.groupBox_3)
        self.width.setSingleStep(0.01)
        self.width.setObjectName("width")
        self.gridLayout_7.addWidget(self.width, 3, 2, 1, 1)
        self.txtCol = QtWidgets.QLineEdit(self.groupBox_3)
        self.txtCol.setObjectName("txtCol")
        self.gridLayout_7.addWidget(self.txtCol, 1, 2, 1, 1)
        self.ndigits = QtWidgets.QSpinBox(self.groupBox_3)
        self.ndigits.setMinimum(1)
        self.ndigits.setMaximum(10)
        self.ndigits.setProperty("value", 1)
        self.ndigits.setObjectName("ndigits")
        self.gridLayout_7.addWidget(self.ndigits, 2, 2, 1, 1)
        self.line_9 = QtWidgets.QFrame(self.groupBox_3)
        self.line_9.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_9.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_9.setObjectName("line_9")
        self.gridLayout_7.addWidget(self.line_9, 2, 1, 1, 1)
        self.label_9 = QtWidgets.QLabel(self.groupBox_3)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_9.setFont(font)
        self.label_9.setObjectName("label_9")
        self.gridLayout_7.addWidget(self.label_9, 1, 0, 1, 1)
        self.label_16 = QtWidgets.QLabel(self.groupBox_3)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_16.setFont(font)
        self.label_16.setObjectName("label_16")
        self.gridLayout_7.addWidget(self.label_16, 2, 0, 1, 1)
        self.line_10 = QtWidgets.QFrame(self.groupBox_3)
        self.line_10.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_10.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_10.setObjectName("line_10")
        self.gridLayout_7.addWidget(self.line_10, 3, 1, 1, 1)
        self.bckCol = QtWidgets.QLineEdit(self.groupBox_3)
        self.bckCol.setObjectName("bckCol")
        self.gridLayout_7.addWidget(self.bckCol, 0, 2, 1, 1)
        self.label_17 = QtWidgets.QLabel(self.groupBox_3)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_17.setFont(font)
        self.label_17.setObjectName("label_17")
        self.gridLayout_7.addWidget(self.label_17, 3, 0, 1, 1)
        self.line_11 = QtWidgets.QFrame(self.groupBox_3)
        self.line_11.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_11.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_11.setObjectName("line_11")
        self.gridLayout_7.addWidget(self.line_11, 4, 1, 1, 1)
        self.label_8 = QtWidgets.QLabel(self.groupBox_3)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_8.setFont(font)
        self.label_8.setObjectName("label_8")
        self.gridLayout_7.addWidget(self.label_8, 0, 0, 1, 1)
        self.line_6 = QtWidgets.QFrame(self.groupBox_3)
        self.line_6.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_6.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_6.setObjectName("line_6")
        self.gridLayout_7.addWidget(self.line_6, 0, 1, 1, 1)
        self.verticalLayout_9.addLayout(self.gridLayout_7)
        self.CbarLayout.addWidget(self.groupBox_3)
        self.groupBox_2 = QtWidgets.QGroupBox(self._cbar_grp)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_4.setContentsMargins(0, -1, 0, -1)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.cmapRev = QtWidgets.QCheckBox(self.groupBox_2)
        self.cmapRev.setObjectName("cmapRev")
        self.gridLayout.addWidget(self.cmapRev, 1, 3, 1, 1)
        self.line_2 = QtWidgets.QFrame(self.groupBox_2)
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.gridLayout.addWidget(self.line_2, 4, 1, 1, 1)
        self.vmaxW = QtWidgets.QWidget(self.groupBox_2)
        self.vmaxW.setObjectName("vmaxW")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.vmaxW)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.gridLayout_3 = QtWidgets.QGridLayout()
        self.gridLayout_3.setVerticalSpacing(9)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.over = QtWidgets.QLineEdit(self.vmaxW)
        self.over.setObjectName("over")
        self.gridLayout_3.addWidget(self.over, 1, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.vmaxW)
        self.label_5.setObjectName("label_5")
        self.gridLayout_3.addWidget(self.label_5, 0, 0, 1, 1)
        self.vmax = QtWidgets.QDoubleSpinBox(self.vmaxW)
        self.vmax.setObjectName("vmax")
        self.gridLayout_3.addWidget(self.vmax, 0, 1, 1, 1)
        self.label_6 = QtWidgets.QLabel(self.vmaxW)
        self.label_6.setObjectName("label_6")
        self.gridLayout_3.addWidget(self.label_6, 1, 0, 1, 1)
        self.verticalLayout_3.addLayout(self.gridLayout_3)
        self.gridLayout.addWidget(self.vmaxW, 4, 2, 1, 2)
        self.label_2 = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 1)
        self.line_4 = QtWidgets.QFrame(self.groupBox_2)
        self.line_4.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.gridLayout.addWidget(self.line_4, 1, 1, 1, 1)
        self.label = QtWidgets.QLabel(self.groupBox_2)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label.setFont(font)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)
        self.isvmin = QtWidgets.QCheckBox(self.groupBox_2)
        font = QtGui.QFont()
        font.setItalic(True)
        self.isvmin.setFont(font)
        self.isvmin.setObjectName("isvmin")
        self.gridLayout.addWidget(self.isvmin, 3, 0, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 0, 2, 1, 1)
        self.cmap = QtWidgets.QComboBox(self.groupBox_2)
        self.cmap.setObjectName("cmap")
        self.gridLayout.addWidget(self.cmap, 1, 2, 1, 1)
        self.widget = QtWidgets.QWidget(self.groupBox_2)
        self.widget.setObjectName("widget")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.widget)
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.climm = QtWidgets.QDoubleSpinBox(self.widget)
        self.climm.setMinimum(-16777215.0)
        self.climm.setMaximum(16777215.0)
        self.climm.setObjectName("climm")
        self.horizontalLayout_3.addWidget(self.climm)
        self.climM = QtWidgets.QDoubleSpinBox(self.widget)
        self.climM.setMinimum(-16777215.0)
        self.climM.setMaximum(16777215.0)
        self.climM.setProperty("value", 1.0)
        self.climM.setObjectName("climM")
        self.horizontalLayout_3.addWidget(self.climM)
        self.horizontalLayout_4.addLayout(self.horizontalLayout_3)
        self.gridLayout.addWidget(self.widget, 2, 2, 1, 2)
        self.line_3 = QtWidgets.QFrame(self.groupBox_2)
        self.line_3.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.gridLayout.addWidget(self.line_3, 2, 1, 1, 1)
        self.limTxt = QtWidgets.QCheckBox(self.groupBox_2)
        self.limTxt.setChecked(True)
        self.limTxt.setObjectName("limTxt")
        self.gridLayout.addWidget(self.limTxt, 5, 0, 1, 4)
        self.line = QtWidgets.QFrame(self.groupBox_2)
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 3, 1, 1, 1)
        self.vminW = QtWidgets.QWidget(self.groupBox_2)
        self.vminW.setObjectName("vminW")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.vminW)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.gridLayout_2 = QtWidgets.QGridLayout()
        self.gridLayout_2.setVerticalSpacing(9)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_3 = QtWidgets.QLabel(self.vminW)
        self.label_3.setObjectName("label_3")
        self.gridLayout_2.addWidget(self.label_3, 0, 0, 1, 1)
        self.vmin = QtWidgets.QDoubleSpinBox(self.vminW)
        self.vmin.setDecimals(3)
        self.vmin.setObjectName("vmin")
        self.gridLayout_2.addWidget(self.vmin, 0, 1, 1, 1)
        self.label_4 = QtWidgets.QLabel(self.vminW)
        self.label_4.setObjectName("label_4")
        self.gridLayout_2.addWidget(self.label_4, 1, 0, 1, 1)
        self.under = QtWidgets.QLineEdit(self.vminW)
        self.under.setObjectName("under")
        self.gridLayout_2.addWidget(self.under, 1, 1, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout_2)
        self.gridLayout.addWidget(self.vminW, 3, 2, 1, 2)
        self.isvmax = QtWidgets.QCheckBox(self.groupBox_2)
        font = QtGui.QFont()
        font.setItalic(True)
        self.isvmax.setFont(font)
        self.isvmax.setObjectName("isvmax")
        self.gridLayout.addWidget(self.isvmax, 4, 0, 1, 1)
        self.widget_2 = QtWidgets.QWidget(self.groupBox_2)
        self.widget_2.setMinimumSize(QtCore.QSize(0, 0))
        self.widget_2.setObjectName("widget_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.widget_2)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.autoscale = QtWidgets.QPushButton(self.widget_2)
        self.autoscale.setObjectName("autoscale")
        self.horizontalLayout.addWidget(self.autoscale)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem3)
        self.gridLayout.addWidget(self.widget_2, 6, 0, 1, 4)
        self.verticalLayout_4.addLayout(self.gridLayout)
        self.CbarLayout.addWidget(self.groupBox_2)
        self.groupBox = QtWidgets.QGroupBox(self._cbar_grp)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_5.setContentsMargins(0, -1, 0, -1)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.gridLayout_4 = QtWidgets.QGridLayout()
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.line_5 = QtWidgets.QFrame(self.groupBox)
        self.line_5.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_5.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_5.setObjectName("line_5")
        self.gridLayout_4.addWidget(self.line_5, 0, 1, 1, 1)
        self.widget_3 = QtWidgets.QWidget(self.groupBox)
        self.widget_3.setObjectName("widget_3")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.widget_3)
        self.verticalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.gridLayout_5 = QtWidgets.QGridLayout()
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.cblabel = QtWidgets.QLineEdit(self.widget_3)
        self.cblabel.setObjectName("cblabel")
        self.gridLayout_5.addWidget(self.cblabel, 0, 1, 1, 1)
        self.label_11 = QtWidgets.QLabel(self.widget_3)
        self.label_11.setObjectName("label_11")
        self.gridLayout_5.addWidget(self.label_11, 1, 0, 1, 1)
        self.label_10 = QtWidgets.QLabel(self.widget_3)
        self.label_10.setObjectName("label_10")
        self.gridLayout_5.addWidget(self.label_10, 0, 0, 1, 1)
        self.cbTxtSz = QtWidgets.QDoubleSpinBox(self.widget_3)
        self.cbTxtSz.setDecimals(1)
        self.cbTxtSz.setObjectName("cbTxtSz")
        self.gridLayout_5.addWidget(self.cbTxtSz, 1, 1, 1, 1)
        self.label_13 = QtWidgets.QLabel(self.widget_3)
        self.label_13.setObjectName("label_13")
        self.gridLayout_5.addWidget(self.label_13, 2, 0, 1, 1)
        self.cbTxtSh = QtWidgets.QDoubleSpinBox(self.widget_3)
        self.cbTxtSh.setDecimals(1)
        self.cbTxtSh.setSingleStep(0.1)
        self.cbTxtSh.setObjectName("cbTxtSh")
        self.gridLayout_5.addWidget(self.cbTxtSh, 2, 1, 1, 1)
        self.verticalLayout_6.addLayout(self.gridLayout_5)
        self.gridLayout_4.addWidget(self.widget_3, 0, 2, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_7.setFont(font)
        self.label_7.setObjectName("label_7")
        self.gridLayout_4.addWidget(self.label_7, 0, 0, 1, 1)
        self.label_12 = QtWidgets.QLabel(self.groupBox)
        font = QtGui.QFont()
        font.setItalic(True)
        self.label_12.setFont(font)
        self.label_12.setObjectName("label_12")
        self.gridLayout_4.addWidget(self.label_12, 1, 0, 1, 1)
        self.line_8 = QtWidgets.QFrame(self.groupBox)
        self.line_8.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_8.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_8.setObjectName("line_8")
        self.gridLayout_4.addWidget(self.line_8, 1, 1, 1, 1)
        self.widget_4 = QtWidgets.QWidget(self.groupBox)
        self.widget_4.setObjectName("widget_4")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.widget_4)
        self.verticalLayout_8.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_8.setSpacing(0)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.gridLayout_6 = QtWidgets.QGridLayout()
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.txtSh = QtWidgets.QDoubleSpinBox(self.widget_4)
        self.txtSh.setDecimals(1)
        self.txtSh.setSingleStep(0.1)
        self.txtSh.setObjectName("txtSh")
        self.gridLayout_6.addWidget(self.txtSh, 1, 1, 1, 1)
        self.txtSz = QtWidgets.QDoubleSpinBox(self.widget_4)
        self.txtSz.setDecimals(1)
        self.txtSz.setObjectName("txtSz")
        self.gridLayout_6.addWidget(self.txtSz, 0, 1, 1, 1)
        self.label_14 = QtWidgets.QLabel(self.widget_4)
        self.label_14.setObjectName("label_14")
        self.gridLayout_6.addWidget(self.label_14, 0, 0, 1, 1)
        self.label_15 = QtWidgets.QLabel(self.widget_4)
        self.label_15.setObjectName("label_15")
        self.gridLayout_6.addWidget(self.label_15, 1, 0, 1, 1)
        spacerItem4 = QtWidgets.QSpacerItem(40, 0, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout_6.addItem(spacerItem4, 2, 1, 1, 1)
        self.verticalLayout_8.addLayout(self.gridLayout_6)
        self.gridLayout_4.addWidget(self.widget_4, 1, 2, 1, 1)
        self.verticalLayout_5.addLayout(self.gridLayout_4)
        spacerItem5 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_5.addItem(spacerItem5)
        self.CbarLayout.addWidget(self.groupBox)
        self.verticalLayout_10.addLayout(self.CbarLayout)
        self.verticalLayout.addWidget(self._cbar_grp)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self._cbar_grp.setTitle(_translate("Form", "Display"))
        self.groupBox_4.setTitle(_translate("Form", "Objects"))
        self.object.setToolTip(_translate("Form", "<html><head/><body><p>Colorbar object to control</p></body></html>"))
        self.label_18.setText(_translate("Form", "Colorbar of"))
        self.groupBox_3.setTitle(_translate("Form", "Settings"))
        self.border.setToolTip(_translate("Form", "<html><head/><body><p>Display / hide the border of the colorbar</p></body></html>"))
        self.border.setText(_translate("Form", "Border"))
        self.bw.setToolTip(_translate("Form", "<html><head/><body><p>Width of the border</p></body></html>"))
        self.width.setToolTip(_translate("Form", "<html><head/><body><p>Width of the colorbar</p></body></html>"))
        self.txtCol.setToolTip(_translate("Form", "<html><head/><body><p>Color of the text and of the border (if shown)</p></body></html>"))
        self.ndigits.setToolTip(_translate("Form", "<html><head/><body><p>Number of digits a fine control of clim/vmin/vmax</p></body></html>"))
        self.label_9.setText(_translate("Form", "Textcolor"))
        self.label_16.setText(_translate("Form", "Number of\n"
"digits"))
        self.bckCol.setToolTip(_translate("Form", "<html><head/><body><p>Background color of the colorbar canvas</p></body></html>"))
        self.label_17.setText(_translate("Form", "Colorbar\n"
"width"))
        self.label_8.setText(_translate("Form", "Background\n"
"color"))
        self.groupBox_2.setTitle(_translate("Form", "Color properties"))
        self.cmapRev.setToolTip(_translate("Form", "<html><head/><body><p>Reverse the colormap\'s colors</p></body></html>"))
        self.cmapRev.setText(_translate("Form", "Reversed"))
        self.over.setToolTip(_translate("Form", "<html><head/><body><p>Every value under vmax are going to be set to this color</p></body></html>"))
        self.label_5.setText(_translate("Form", "Value"))
        self.vmax.setToolTip(_translate("Form", "<html><head/><body><p>Value of the maximum threshold</p></body></html>"))
        self.label_6.setText(_translate("Form", "Over\n"
"color"))
        self.label_2.setText(_translate("Form", "Clim"))
        self.label.setText(_translate("Form", "Colormap"))
        self.isvmin.setToolTip(_translate("Form", "<html><head/><body><p>Enable / Disable a minimum threshold</p></body></html>"))
        self.isvmin.setText(_translate("Form", "Vmin"))
        self.cmap.setToolTip(_translate("Form", "<html><head/><body><p>Colormap to use</p></body></html>"))
        self.climm.setToolTip(_translate("Form", "<html><head/><body><p>Minimum of the colorbar</p></body></html>"))
        self.climM.setToolTip(_translate("Form", "<html><head/><body><p>Maximum of the colorbar</p></body></html>"))
        self.limTxt.setToolTip(_translate("Form", "<html><head/><body><p>Display vmin/vmax on the colorbar</p></body></html>"))
        self.limTxt.setText(_translate("Form", "Display vmin/vmax text"))
        self.label_3.setText(_translate("Form", "Value"))
        self.vmin.setToolTip(_translate("Form", "<html><head/><body><p>Value of the minimum threshold</p></body></html>"))
        self.label_4.setText(_translate("Form", "Under\n"
"color"))
        self.under.setToolTip(_translate("Form", "<html><head/><body><p>Every values over vmin are going to be set to this color</p></body></html>"))
        self.isvmax.setToolTip(_translate("Form", "<html><head/><body><p>Enable / Disable a maximum threshold</p></body></html>"))
        self.isvmax.setText(_translate("Form", "Vmax"))
        self.autoscale.setToolTip(_translate("Form", "<html><head/><body><p>Automatically scale the colorbar limits to fit with the (minimum, maximum) of currently displayed object</p></body></html>"))
        self.autoscale.setText(_translate("Form", "Autoscale"))
        self.groupBox.setTitle(_translate("Form", "Text properties"))
        self.cblabel.setToolTip(_translate("Form", "<html><head/><body><p>Title of the colorbar</p></body></html>"))
        self.label_11.setText(_translate("Form", "Size"))
        self.label_10.setText(_translate("Form", "Value"))
        self.cbTxtSz.setToolTip(_translate("Form", "<html><head/><body><p>Text size of the colorbar label</p></body></html>"))
        self.label_13.setText(_translate("Form", "Shift"))
        self.cbTxtSh.setToolTip(_translate("Form", "<html><head/><body><p>Shift between the colorbar and the text</p></body></html>"))
        self.label_7.setText(_translate("Form", "Title"))
        self.label_12.setText(_translate("Form", "Limits"))
        self.txtSh.setToolTip(_translate("Form", "<html><head/><body><p>Shift between the colorbar and limits (clim / vmin / vmax)</p></body></html>"))
        self.txtSz.setToolTip(_translate("Form", "<html><head/><body><p>Text size of limits (clim / vmin / vmax)</p></body></html>"))
        self.label_14.setText(_translate("Form", "Size"))
        self.label_15.setText(_translate("Form", "Shift"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())




"""Surfaces (mesh) and volume utility functions."""
import logging

import numpy as np
from scipy.spatial.distance import cdist

from vispy.geometry import MeshData
from vispy.geometry.isosurface import isosurface

from .sigproc import smooth_3d


__all__ = ('vispy_array', 'convert_meshdata', 'volume_to_mesh',
           'smoothing_matrix', 'mesh_edges', 'laplacian_smoothing')


logger = logging.getLogger('visbrain')


def vispy_array(data, dtype=np.float32):
    """Check and convert array to be compatible with buffers.

    Parameters
    ----------
    data : array_like
        Array of data.
    dtype : type | np.float32
        Futur type of the array.

    Returns
    -------
    data : array_like
        Contiguous array of type dtype.
    """
    if not data.flags['C_CONTIGUOUS']:
        data = np.ascontiguousarray(data, dtype=dtype)
    if data.dtype != dtype:
        data = data.astype(dtype, copy=False)
    return data


def convert_meshdata(vertices=None, faces=None, normals=None, meshdata=None,
                     invert_normals=False, transform=None):
    """Convert mesh data to be compatible with visbrain.

    Parameters
    ----------
    vertices : array_like | None
        Vertices of the template of shape (N, 3) or (N, 3, 3) if indexed by
        faces.
    faces : array_like | None
        Faces of the template of shape (M, 3)
    normals : array_like | None
        The normals to each vertex, with the same shape as vertices.
    meshdata : VisPy.MeshData | None
        VisPy MeshData object.
    invert_normals : bool | False
        If the brain appear to be black, use this parameter to invert normals.
    transform : visPy.transform | None
        VisPy transformation to apply to vertices ans normals.

    Returns
    -------
    vertices : array_like
        Vertices of shape (N, 3)
    faces : array_like
        Faces of the template of shape (M, 3)
    normals : array_like
        The normals of shape (M, 3, 3).
    """
    # Priority to meshdata :
    if meshdata is not None:
        vertices = meshdata.get_vertices()
        faces = meshdata.get_faces()
        normals = meshdata.get_vertex_normals()
        logger.debug('Indexed faces normals converted // extracted')
    else:
        # Check if faces index start at zero (Matlab like):
        if faces.min() != 0:
            faces -= faces.min()
        # Get normals if None :
        if (normals is None) or (normals.ndim != 2):
            md = MeshData(vertices=vertices, faces=faces)
            normals = md.get_vertex_normals()
            logger.debug('Indexed faces normals converted // extracted')
    assert vertices.ndim == 2

    # Invert normals :
    norm_coef = -1. if invert_normals else 1.
    normals *= norm_coef

    # Apply transformation :
    if transform is not None:
        vertices = transform.map(vertices)[..., 0:-1]
        normals = transform.map(normals)[..., 0:-1]

    # Type checking :
    vertices = vispy_array(vertices)
    faces = vispy_array(faces, np.uint32)
    normals = vispy_array(normals)

    return vertices, faces, normals


def volume_to_mesh(vol, smooth_factor=3, level=None, **kwargs):
    """Convert a volume into a mesh with vertices, faces and normals.

    Parameters
    ----------
    vol : array_like
        The volume of shape (N, M, P)
    smooth_factor : int | 3
        The smoothing factor to apply to the volume.
    level : int | None
        Level to extract.
    kwargs : dict | {}
        Optional arguments to pass to convert_meshdata.

    Returns
    -------
    vertices : array_like
        Mesh vertices.
    faces : array_like
        Mesh faces.
    normals : array_like
        Mesh normals.
    """
    # Smooth the volume :
    vol_s, tf = smooth_3d(vol, smooth_factor, correct=True)
    # Extract vertices and faces :
    if level is None:
        level = .5
    elif isinstance(level, int):
        vol_s[vol_s != level] = 0
        level = .5
    vert_n, faces_n = isosurface(vol_s, level=level)
    # Smoothing compensation :
    vert_n = tf.map(vert_n)[:, 0:-1]
    # Convert to meshdata :
    vertices, faces, normals = convert_meshdata(vert_n, faces_n, **kwargs)
    return vertices, faces, normals


def smoothing_matrix(vertices, adj_mat, smoothing_steps=20):
    """Create a smoothing matrix.

    This function  can be used to interpolate data defined for a subset of
    vertices onto mesh with an adjancency matrix given by adj_mat.

    This function is a copy from the PySurfer package. See :
    https://github.com/nipy/PySurfer/blob/master/surfer/utils.py

    Parameters
    ----------
    vertices : array_like
        Vertex indices of shape (N,)
    adj_mat : sparse matrix
        N x N adjacency matrix of the full mesh.
    smoothing_steps : int
        Number of smoothing steps. If smoothing_steps is None, as many
        smoothing steps are applied until the whole mesh is filled with
        with non-zeros. Only use this option if the vertices correspond to a
        subsampled version of the mesh.
    Returns
    -------
    smooth_mat : sparse matrix
        smoothing matrix with size N x len(vertices)
    """
    from scipy import sparse

    e = adj_mat.copy()
    e.data[e.data == 2] = 1
    n_vertices = e.shape[0]
    e = e + sparse.eye(n_vertices, n_vertices)
    idx_use = vertices
    smooth_mat = 1.0
    n_iter = smoothing_steps if smoothing_steps is not None else 1000
    for k in range(n_iter):
        e_use = e[:, idx_use]

        data1 = e_use * np.ones(len(idx_use))
        idx_use = np.where(data1)[0]
        scale_mat = sparse.dia_matrix((1 / data1[idx_use], 0),
                                      shape=(len(idx_use), len(idx_use)))

        smooth_mat = scale_mat * e_use[idx_use, :] * smooth_mat

        if smoothing_steps is None and len(idx_use) >= n_vertices:
            break

    # Make sure the smoothing matrix has the right number of rows
    # and is in COO format
    smooth_mat = smooth_mat.tocoo()
    smooth_mat = sparse.coo_matrix((smooth_mat.data,
                                    (idx_use[smooth_mat.row],
                                     smooth_mat.col)),
                                   shape=(n_vertices,
                                          len(vertices)))

    return smooth_mat


def mesh_edges(faces):
    """Get sparse matrix with edges as an adjacency matrix.

    This function is a copy from the PySurfer package. See :
    https://github.com/nipy/PySurfer/blob/master/surfer/utils.py

    Parameters
    ----------
    faces : array_like
        The mesh faces of shape (n_faces, 3).
    Returns
    -------
    edges : sparse matrix
        The adjacency matrix.
    """
    from scipy import sparse
    npoints = np.max(faces) + 1
    nfaces = len(faces)
    a, b, c = faces.T
    edges = sparse.coo_matrix((np.ones(nfaces), (a, b)),
                              shape=(npoints, npoints))
    edges = edges + sparse.coo_matrix((np.ones(nfaces), (b, c)),
                                      shape=(npoints, npoints))
    edges = edges + sparse.coo_matrix((np.ones(nfaces), (c, a)),
                                      shape=(npoints, npoints))
    edges = edges + edges.T
    edges = edges.tocoo()
    return edges


def laplacian_smoothing(vertices, faces, n_neighbors=-1):
    """Apply a laplacian smoothing to vertices.

    Parameters
    ----------
    vertices : array_like
        Array of vertices.
    vertices : array_like
        Array of faces.
    n_neighbors : int | -1
        Specify maximum number of closest neighbors to take into account in the
        mean.

    Returns
    -------
    new_vertices : array_like
        New smoothed vertices.
    """
    assert vertices.ndim == 2 and vertices.shape[1] == 3
    assert faces.ndim == 2 and faces.shape[1] == 3
    assert n_neighbors >= -1 and isinstance(n_neighbors, int)
    n_vertices = vertices.shape[0]
    new_vertices = np.zeros_like(vertices)
    for k in range(n_vertices):
        # Find connected vertices :
        faces_idx = np.where(faces == k)[0]
        u_faces_idx = np.unique(np.ravel(faces[faces_idx, :])).tolist()
        u_faces_idx.remove(k)
        # Select closest connected vertices :
        if n_neighbors == -1:
            to_smooth = u_faces_idx
        else:
            norms = cdist(vertices[[k], :], vertices[u_faces_idx, :]).ravel()
            n_norm = min(n_neighbors, len(norms))
            to_smooth = np.array(u_faces_idx)[np.argsort(norms)[0:n_norm]]
        # Take the mean of selected vertices :
        new_vertices[k, :] = vertices[to_smooth, :].mean(0).reshape(1, -1)
    return new_vertices



"""Colorbar object."""
import logging

from vispy import scene

from .visbrain_obj import VisbrainObject
from ..visuals import CbarVisual

logger = logging.getLogger('visbrain')


class ColorbarObj(VisbrainObject):
    """Create a colorbar object.

    Parameters
    ----------
    name : str
        Name of the colorbar object. Alternatively, you can pass an other
        object (like BrainObj or SourceObj) to get their colorbar.
    rect : tuple | (-.7, -2, 1.5, 4)
        Camera rectangle. The `rect` input must be a tuple of four floats
        describing where the camera (start_x, start_y, length_x, length_y).
    cmap : string | None
        Matplotlib colormap (like 'viridis', 'inferno'...).
    clim : tuple/list | None
        Colorbar limit. Every values under / over clim will
        clip.
    isvmin : bool | False
        Activate/deactivate vmin.
    vmin : float | None
        Every values under vmin will have the color defined
        using the under parameter.
    vmax : float | None
        Every values over vmin will have the color defined
        using the over parameter.
    under : tuple/string | None
        Matplotlib color under vmin.
    over : tuple/string | None
        Matplotlib color over vmax.
    cblabel : string | ''
        Colorbar label.
    cbtxtsz : float | 5..
        Text size of the colorbar label.
    cbtxtsh : float | 2.3
        Shift for the colorbar label.
    txtcolor : string | 'white'
        Text color.
    txtsz : float | 3.
        Text size for clim/vmin/vmax text.
    txtsh : float | 1.2
        Shift for clim/vmin/vmax text.
    border : bool | True
        Display colorbar borders.
    bw : float | 2.
        Border width.
    limtxt : bool | True
        Display vmin/vmax text.
    bgcolor : tuple/string | (0., 0., 0.)
        Background color of the colorbar canvas.
    ndigits : int | 2
        Number of digits for the text.
    width : float | 0.17
        Colorbar width.
    transform : VisPy.visuals.transforms | None
        VisPy transformation to set to the parent node.
    parent : VisPy.parent | None
        Markers object parent.
    verbose : string
        Verbosity level.

    Notes
    -----
    List of supported shortcuts :

        * **s** : save the figure
        * **<delete>** : reset camera

    Examples
    --------
    >>> from sleep_scoring.objects import ColorbarObj
    >>> cb = ColorbarObj('cbar', cmap='viridis', clim=(4., 78.2), vmin=10.,
    >>>                  vmax=72., cblabel='Colorbar title', under='gray',
    >>>                  over='red', txtcolor='black', cbtxtsz=40, cbtxtsh=2.,
    >>>                  txtsz=20., width=.04)
    >>> cb.preview()
    """

    def __init__(self, name, rect=(-.7, -2, 1.5, 4), transform=None,
                 parent=None, verbose=None, **kwargs):
        """Init."""
        # Init Visbrain object base class and SourceProjection :
        if not isinstance(name, str):
            kwargs = self._update_cbar_from_obj(name, update=False, **kwargs)
            if hasattr(name, '_default_cblabel') and (
                    'cblabel' not in kwargs.keys()):
                kwargs['cblabel'] = name._default_cblabel
            name = name.name + 'Cbar'  # that's a lot of name
        kwargs['isvmin'] = kwargs.get('vmin', None) is not None
        kwargs['isvmax'] = kwargs.get('vmax', None) is not None
        VisbrainObject.__init__(self, name, parent, transform, verbose)
        self._cbar = CbarVisual(parent=self._node, **kwargs)
        assert len(rect) == 4
        self._rect = rect

    def _get_camera(self):
        """Get a panzoom camera."""
        return scene.cameras.PanZoomCamera(rect=self._rect)

    def _update_cbar_from_obj(self, obj, update=True, **kwargs):
        """Update colorbar from an object."""
        is_meth = hasattr(obj, 'to_dict') and callable(obj.to_dict)
        if is_meth:
            logger.info("Get colorbar properties from %s object" % repr(obj))
            # Get object cbar properties :
            kw = obj.to_dict()
            if update:  # Update the colorbar visual
                for name, val in kw.items():
                    exec('self._cbar._%s = val' % name)
                self._cbar._build()
            for name, val in kwargs.items():
                kw[name] = val
            return kw
        else:
            raise ValueError("Can not get the colorbar of a %s "
                             "object" % type(obj).__name__)
